#include <vector>
#include <string>

#include "lexeme_impl.h"
#include "function_callback.h"

#include "literal_data.h"

using namespace utakata::lexeme;
using namespace utakata::utf8_string;
using namespace utakata::literal;

////////////////////////
// 各lexeme生成関数   //
////////////////////////
smart_ptr<ILexeme> utakata::lexeme::makeOpenParen()
{
    smart_ptr<UTF8String> tmp(new UTF8String(convert("(")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp, LexemeID::openParenthesis));
}

smart_ptr<ILexeme> utakata::lexeme::makeEOS()
{
    return smart_ptr<ILexeme>(new PureLexeme(smart_ptr<UTF8String>(), LexemeID::eos));
}

smart_ptr<ILexeme> utakata::lexeme::makeCloseParen()
{
    smart_ptr<UTF8String> tmp(new UTF8String(utf8_string::convert(")")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::closeParenthesis));
}

smart_ptr<ILexeme> utakata::lexeme::makeBackQuote()
{
    smart_ptr<UTF8String> tmp(new UTF8String(utf8_string::convert("`")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::backquote));
}

smart_ptr<ILexeme> utakata::lexeme::makeQuote()
{
    smart_ptr<UTF8String> tmp(new UTF8String(utf8_string::convert("'")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::quote));
}

smart_ptr<ILexeme> utakata::lexeme::makeDot()
{
    smart_ptr<UTF8String> tmp(new UTF8String(utf8_string::convert(".")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::dot));
}

smart_ptr<ILexeme> utakata::lexeme::makeIdentifier(const utakata::utf8_string::UTF8String& str)
{
    return smart_ptr<ILexeme>(new Identifier(str));
}

smart_ptr<ILexeme> utakata::lexeme::makeString(const utakata::utf8_string::UTF8String& str)
{
    return smart_ptr<ILexeme>(new String(str));
}

smart_ptr<ILexeme> utakata::lexeme::makeUnquoteSplicing()
{
    smart_ptr<UTF8String> tmp(new UTF8String(utf8_string::convert(",@")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::unquoteSplicing));
}

smart_ptr<ILexeme> utakata::lexeme::makeUnquote()
{
    smart_ptr<UTF8String> tmp(new UTF8String(utf8_string::convert(",")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::unquote));
}

smart_ptr<ILexeme> utakata::lexeme::makeByteVector()
{
    smart_ptr<UTF8String> tmp(new UTF8String(utf8_string::convert("#vu(")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::byteVector));
}

smart_ptr<ILexeme> utakata::lexeme::makeVector()
{
    smart_ptr<UTF8String> tmp(new UTF8String(utf8_string::convert("#(")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::vector));
}

smart_ptr<ILexeme> utakata::lexeme::makeSyntax()
{
    smart_ptr<UTF8String> tmp(new UTF8String(utf8_string::convert("#'")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::syntax));
}

smart_ptr<ILexeme> utakata::lexeme::makeQuasiSyntax()
{
    smart_ptr<UTF8String> tmp(new UTF8String(utf8_string::convert("#`")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::quasiSyntax));
}

smart_ptr<ILexeme> utakata::lexeme::makeUnsyntaxSplicing()
{
    smart_ptr<UTF8String> tmp(new UTF8String(utf8_string::convert("#,@")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::unsyntaxSplicing));
}

smart_ptr<ILexeme> utakata::lexeme::makeUnsyntax()
{
    smart_ptr<UTF8String> tmp(new UTF8String(utf8_string::convert("#,")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::unsyntax));
}

smart_ptr<ILexeme> utakata::lexeme::makeCharactor(const utakata::utf8_string::UTF8String& str)
{
    return smart_ptr<ILexeme>(new Charactor(str));
}

smart_ptr<ILexeme> utakata::lexeme::makeBoolean(smart_ptr<utakata::utf8_string::UTF8Char> ch)
{
    return smart_ptr<ILexeme>(new Boolean(ch));
}

smart_ptr<ILexeme> utakata::lexeme::makeNanImaginary(const utakata::utf8_string::UTF8String& str,
                                                     bool exact)
{
    return smart_ptr<ILexeme>(new Number(utakata::utf8_string::UTF8String(),
                                         str, exact, 10));
}

smart_ptr<ILexeme> utakata::lexeme::makeInfImaginary(const utakata::utf8_string::UTF8String& str,
                                                     bool exact)
{
    return smart_ptr<ILexeme>(new Number(utakata::utf8_string::UTF8String(),
                                         str, exact, 10));
}

smart_ptr<ILexeme> utakata::lexeme::makeImaginaryOnly(const utakata::utf8_string::UTF8String& str,
                                                      bool exact)
{

    return smart_ptr<ILexeme>(new Number(utakata::utf8_string::UTF8String(),
                                         str, exact, 10));
}

smart_ptr<ILexeme> utakata::lexeme::makeNumber(const utakata::utf8_string::UTF8String& real,
                                               const utakata::utf8_string::UTF8String& imagin,
                                               bool exact, int radix)
{
    return smart_ptr<ILexeme>(new Number(real, imagin, exact, radix));
}

////////////////
// PureLexeme //
////////////////

PureLexeme::PureLexeme(smart_ptr<UTF8String> str,
                       lexeme::LexemeID id) : str_(str), id_(id)
{}

const LexemeID PureLexeme::getID() const
{
    return id_;
}

smart_ptr<UTF8String> PureLexeme::toString() const
{
    return str_;
}

smart_ptr<utakata::literal::LiteralData> PureLexeme::getData() const
{
    return smart_ptr<utakata::literal::LiteralData>();
}

////////////////
// Identifier //
////////////////

Identifier::Identifier(const utakata::utf8_string::UTF8String& str) :
    str_(new utakata::utf8_string::UTF8String(str))
{
}

const utakata::lexeme::LexemeID Identifier::getID() const
{
    return utakata::lexeme::LexemeID::identifier;
}

smart_ptr<utakata::utf8_string::UTF8String> Identifier::toString() const
{
    return str_;
}

smart_ptr<utakata::literal::LiteralData> Identifier::getData() const
{
    return utakata::literal::makeSymbol(*str_);
}

/////////////
// String  //
/////////////

String::String(const utakata::utf8_string::UTF8String& str) :
    str_(new utakata::utf8_string::UTF8String(str))
{
}

const utakata::lexeme::LexemeID String::getID() const
{
    return utakata::lexeme::LexemeID::string;
}

smart_ptr<utakata::utf8_string::UTF8String> String::toString() const
{
    return str_;
}

smart_ptr<utakata::literal::LiteralData> String::getData() const
{
    return utakata::literal::makeString(*str_);
}

////////////
// Number //
////////////

Number::Number(const UTF8String& real,
               const UTF8String& imagin,
               bool exact, int radix) :
    data_(new LiteralData)
{
    data_ = utakata::literal::makeNumber(real, imagin, exact, radix);
}

const utakata::lexeme::LexemeID Number::getID() const
{
    return utakata::lexeme::LexemeID::number;
}

smart_ptr<utakata::utf8_string::UTF8String> Number::toString() const
{
    return smart_ptr<UTF8String>(new UTF8String(*data_->number->real + *data_->number->imagin));
}

smart_ptr<utakata::literal::LiteralData> Number::getData() const
{
    return data_;
}

///////////////
// Charactor //
///////////////

Charactor::Charactor(const UTF8String& ch) : ch_(new UTF8String(ch))
{
}

const utakata::lexeme::LexemeID Charactor::getID() const
{
    return utakata::lexeme::LexemeID::charactor;
}

smart_ptr<utakata::utf8_string::UTF8String> Charactor::toString() const
{
    return ch_;
}

smart_ptr<utakata::literal::LiteralData> Charactor::getData() const
{
    return literal::makeCharactor(*ch_);
}

/////////////
// Boolean //
/////////////

Boolean::Boolean(smart_ptr<UTF8Char> ch) : bool_(new UTF8String())
{
    *bool_ += "#";
    *bool_ += *ch;
}

const utakata::lexeme::LexemeID Boolean::getID() const
{
    return utakata::lexeme::LexemeID::boolean;
}

smart_ptr<utakata::utf8_string::UTF8String> Boolean::toString() const
{
    return bool_;
}

smart_ptr<utakata::literal::LiteralData> Boolean::getData() const
{
    return utakata::literal::makeBoolean(*bool_, true);
}
