#ifndef _LEXEME_IMPL_H_
#define _LEXEME_IMPL_H_

#include <string>

#include "lexeme.h"
#include "smart_ptr.h"
#include "function_callback.h"

#include "utf8_string.h"
#include "lexeme_id.h"

// lexeme の様々な実装を定義する.
namespace utakata {

    namespace lexeme {
        class LexemeID;

        // 各 lexeme を smart_ptr にして取得する.
        smart_ptr<ILexeme> makeOpenParen();
        smart_ptr<ILexeme> makeEOS();
        smart_ptr<ILexeme> makeCloseParen();
        smart_ptr<ILexeme> makeBackQuote ();
        smart_ptr<ILexeme> makeQuote ();
        smart_ptr<ILexeme> makeDot ();
        smart_ptr<ILexeme> makeUnquoteSplicing ();
        smart_ptr<ILexeme> makeUnquote ();
        smart_ptr<ILexeme> makeSyntax ();
        smart_ptr<ILexeme> makeQuasiSyntax ();
        smart_ptr<ILexeme> makeUnsyntaxSplicing ();
        smart_ptr<ILexeme> makeUnsyntax ();
        smart_ptr<ILexeme> makeByteVector ();
        smart_ptr<ILexeme> makeVector ();
        smart_ptr<ILexeme> makeIdentifier (const utakata::utf8_string::UTF8String& str);
        smart_ptr<ILexeme> makeString (const utakata::utf8_string::UTF8String& str);
        smart_ptr<ILexeme> makeNumber (const utakata::utf8_string::UTF8String& real,
                                      const utakata::utf8_string::UTF8String& imaginary,
                                      bool exact, int radix);
        smart_ptr<ILexeme> makeCharactor (const utakata::utf8_string::UTF8String& str);
        smart_ptr<ILexeme> makeNanImaginary (const utakata::utf8_string::UTF8String& str,
                                            bool exact);
        smart_ptr<ILexeme> makeInfImaginary (const utakata::utf8_string::UTF8String& str,
                                            bool exact);
        smart_ptr<ILexeme> makeImaginaryOnly (const utakata::utf8_string::UTF8String& str,
                                             bool exact);
        smart_ptr<ILexeme> makeBoolean (smart_ptr<utakata::utf8_string::UTF8Char> ch);


        ////////////////////////////////////////////
        // 実際に利用する lexeme は以下に実装する. //
        ////////////////////////////////////////////

        class PureLexeme : public ILexeme
        {
            // LiteralDataを必要としないデータに対して処理を行うためのクラス。
            // 文字列と該当するLexemeIDを渡すことで実現する。
        public:
            PureLexeme(smart_ptr<utakata::utf8_string::UTF8String> str,
                       lexeme::LexemeID id);

            virtual ~PureLexeme() {}

            const lexeme::LexemeID getID() const ;
            smart_ptr<utakata::utf8_string::UTF8String> toString() const;
            smart_ptr<utakata::literal::LiteralData> getData() const;

        private:

            smart_ptr<utakata::utf8_string::UTF8String> str_;
            lexeme::LexemeID id_;
        };

        ////////////
        // String //
        ////////////

        class String : public ILexeme
        {
        public:
            
            String (const utakata::utf8_string::UTF8String& ident);
            virtual ~String (){}
            const lexeme::LexemeID getID () const;
            smart_ptr<utakata::utf8_string::UTF8String> toString () const;
            smart_ptr<utakata::literal::LiteralData> getData() const;

        private:

            smart_ptr<utakata::utf8_string::UTF8String> str_;

        };

        ////////////////
        // Identifier //
        ////////////////

        class Identifier : public ILexeme
        {
        public:
            
            Identifier (const utakata::utf8_string::UTF8String& ident);
            virtual ~Identifier (){}
            const lexeme::LexemeID getID () const;
            smart_ptr<utakata::utf8_string::UTF8String> toString () const;
            smart_ptr<utakata::literal::LiteralData> getData() const;
            
        private:

            smart_ptr<utakata::utf8_string::UTF8String> str_;

        };

        ////////////
        // Number //
        ////////////

        class Number : public ILexeme
        {
        public:
            Number (const utf8_string::UTF8String& real, const utf8_string::UTF8String& imagin,
                   bool exact, int radix);
            virtual ~Number (){}
            const lexeme::LexemeID getID () const;
            smart_ptr<utakata::utf8_string::UTF8String> toString () const;
            smart_ptr<utakata::literal::LiteralData> getData() const;
        private:

            smart_ptr<utakata::literal::LiteralData> data_;
        };

        ///////////////
        // Charactor //
        ///////////////

        class Charactor : public ILexeme
        {
        public:
            Charactor (const utf8_string::UTF8String& ch);
            virtual ~Charactor (){}

            const lexeme::LexemeID getID () const;
            smart_ptr<utakata::utf8_string::UTF8String> toString() const;
            smart_ptr<utakata::literal::LiteralData> getData() const;

        private:

            smart_ptr<utakata::utf8_string::UTF8String> ch_;
        };

        /////////////
        // Boolean //
        /////////////

        class Boolean : public ILexeme
        {
        public:
            Boolean (smart_ptr<utf8_string::UTF8Char> ch);
            virtual ~Boolean (){}

            const lexeme::LexemeID getID () const;
            smart_ptr<utakata::utf8_string::UTF8String> toString () const;
            smart_ptr<utakata::literal::LiteralData> getData() const;

        private:
            smart_ptr<utf8_string::UTF8String> bool_;
        };

    };

};

#endif /* _LEXEME_IMPL_H_ */
