#ifndef _PM_H_
#define _PM_H_

#include <vector>
#include <algorithm>

namespace vm {

    class multiple_precision_integer
    {
        /**
           多倍長演算を行うためのクラス。
           多倍長整数演算を実装する。

           ここで実装する多倍長整数演算は、特に何も考えないで、次のように単純に実装する。

           桁の単位となる符号無し整数型 r
           整数型のbit数 b

           があった場合、

           多倍長整数の内部表現としては、次のようになる。

           mp = (r[n] <<（b * n - 1)) + (r[n-1] << (b * n - 2))...r[0]

           つまり、各桁の単位となるデータ単位のbitを繋げていくだけ、というシンプルな実装になる。

           負数の表現は、内部変数negate_がfalseである場合は自然数、trueである場合には負数となる。
        */
    private:

        template<class S>
        class nega_bits
        {
        public:
            nega_bits(){}
            virtual ~nega_bits(){}

            S operator()(const S& s) {
                return !s;
            }
        };

    public:

        // 多倍長整数の初期化を行う。
        multiple_precision_integer() : negate_(false), digits_()
            {
                assign(0);
            }

        template<class T>
        multiple_precision_integer(T init) : negate_(false), digits_()
            {
                assign(init);
            }

        multiple_precision_integer(const multiple_precision_integer& init) :
            negate_(false), digits_()
            {
                this->digits_ = init.digits_;
            }
        
        virtual ~multiple_precision_integer() {}

        /**
           多倍長整数に値を設定する。
           多倍長整数であるか、もしくはそれ以外であるかで処理が分岐する。
        */
        template<class S>
        void assign(S dat)
            {
                this->itompi_(dat);
            }

        void assign(const multiple_precision_integer& dat)
            {
                *this = dat;
            }

        /**
           オペレータオーバーロードを行う。
        */
        multiple_precision_integer& operator=(const multiple_precision_integer& s)
            {
                multiple_precision_integer m(s);
                std::swap(m, *this);
                return *this;
            }

        multiple_precision_integer& plus(const multiple_precision_integer& t)
            {
                // 多倍長整数同士の加算を行う。
                // 多倍長整数同士の場合、4パターンの加算パターンが存在し、このうち
                // 2パターンで、加算時に桁溢れが発生する。

                // それぞれを同じサイズにする。
                std::vector<unsigned char> tmp = t.digits_;
                if (digits_.size() < tmp.size())
                {
                    digits_.resize(tmp.size());
                }
                else if (digits_.size() > tmp.size())
                {
                    tmp.resize(digits_.size());
                }

                // 内部の別関数に任せる。
                inner_plus_(tmp, t.negate_);

                return *this;
            }


    private:


        void inner_plus_(const std::vector<unsigned char>& t, bool negate)
            {
                // 自然数、および負数の加算を行う。
                // 実行する前に、すでにtとクラスのvectorは同じサイズにされている。

                std::vector<unsigned char>::iterator beg1 = digits_.begin();
                std::vector<unsigned char>::iterator end1 = digits_.end();
                std::vector<unsigned char>::iterator beg2 = t.begin();
                std::vector<unsigned char>::iterator end2 = t.end();

                if ((negate_ && negate) || (!negate_ && !negate)) {
                    // 両方が負数、もしくは自然数の場合、そのまま加算処理を行う。
                    unsigned char flow = 0;
                    for (;beg1 != end1 && beg2 != end2;beg1++,beg2++)
                    {
                        // 実際に加算する。加算をチェックする場合、上位2bitを加算し、結果をさらに2bit
                        // 右にシフトすることで、flowを取得できる。
                        unsigned char t = ((*beg1 + flow) >> ((sizeof(unsigned char) - 2)) +
                                           (*beg2 >> (sizeof(unsigned char) - 2))) >> 2
                        *beg1 += *beg2 + flow;
                        flow = t;
                    }

                    // 追加させる。
                    if (flow > 0) {
                        digits_.push_back(flow);
                    }
                }
                else
                {
                    // 自身または相手が負数である場合、一度全単位の負数を求める必要がある。

                    
                }

        /**
           整数の表現を多倍長整数の内部表現に変換する。
        */
        void itompi_(T dat)
            {
                //負数であった場合、一度自然数の表現にしてから、再度負数にするという
                //二段階の処理が必要になる.
                
                T maxbit = 1 << (sizeof(T) - 1);
                bool negate = false;
                T tmp = dat;
                if (dat & maxbit)
                {
                    // 負数なので、自然数の表現に変換する。2の補数から自然数表現に戻すには、
                    // 補数表現から1を引いて、結果をnotする。
                    tmp = !(dat - 1);
                    negate = true;
                }

                // 一度値をリセットする。
                init_mp_(tmp, negate);
            }

        void init_mp_(T dat, bool negate)
            {
                // 一度値をリセットする。
                digits_.clear();
                T maxbit = 1 << (sizeof(T) * 8 - 1);

                // 負数かどうかを調べて、負数の場合は格納方法を変える必要がある。
                if (max_bit_ < maxbit) {
                    // 現在保持している値よりも大きい場合

                    for (int i = 0, max = maxbit / max_bit_; i < max;++i) {
                        digits_.push_back(dat & (max_bit_ | (max_bit_ - 1)));
                        dat >>= max_bit_;
                    }

                    // 殘ったビットを追加する。
                    if (maxbit % max_bit_ > 0) {
                        digits_.push_back(dat);
                    }

                } else {
                    // 先頭のbitだけは無視するようにする。
                    digits_.push_back(dat);
                }

                // 負数であるかどうかを調べ、負数である場合には、全体を反転し、最後に1を追加する。
                // 初期化された値が0の場合のみ、ここでの繰り上がりが発生しうる。
                std::transform(digits_.begin(), digits_.end(),
                               digits_.begin(), nega_bits());
                // とりあえず1を加算しておく。
                plus(1);
            }


        /**
           符号無し整数の場合には、下に挙げるオーバーロード関数が利用される。
        */
        void itompi_(unsigned int dat) {init_mp_(dat);}
        void itompi_(unsigned short dat) {init_mp_(dat);}
        void itompi_(unsigned char dat) {init_mp_(dat);}
        void itompi_(unsigned long dat) {init_mp_(dat);}

        // vectorそのものを渡して、結果を直接設定させる。
        void assign_digits(const std::vector<unsigned char>& di)
            {
                this->digits_ = di;
            }

        bool isNegate() const {return negate_;}

    private:

        std::vector<unsigned char> digits_;      // 各桁を設定する。
        bool negate_;                            // 負数であるかどうか。
    };

};

#endif /* _PM_H_ */
