#ifndef _NUMBER_LEXER_H_
#define _NUMBER_LEXER_H_

#include <iostream>

#include <sstream>
#include "utf8_string.h"
#include "smart_ptr.h"
#include "utf8.h"
#include "lexeme.h"
#include "sublexer.h"

namespace utakata {

    namespace sublexer {

        namespace number {
    
            /////////////////////////////////////////////////////////////////
            // <number>がそれぞれのradixに基づいて解析する必要があるため、 //
            // それぞれに対して特化できるようなテンプレートを用意する。    //
            /////////////////////////////////////////////////////////////////

            template<int NUM> class UReal;

            template<int NUM>
            class Number
            {

                const int INIT;
                const int REAL;     // <real>非終端記号を示す。これは決して虚数にはならない。
                const int UREAL;    // <ureal>非終端記号を示す。
                const int NAN_IMAGINARY;
                const int INF_IMAGINARY;
                const int IMAGINARY_ONLY;
                const int FLAG;
                const int END;
                const int IMAGINARY;
            public:
                Number() : 
                        INIT(1), REAL(2), UREAL(3), NAN_IMAGINARY(4), INF_IMAGINARY(5),
                        IMAGINARY_ONLY(8), FLAG(9), END(10), IMAGINARY(11),
                        exact_(false), realnum_(new utf8_string::UTF8String()),
                        imaginary_(new utf8_string::UTF8String())
                {
                }
                
                virtual ~Number() {}

                smart_ptr<lexeme::ILexeme> lex(smart_ptr<utf8::UTF8InputStream> stream,
                                               smart_ptr<ISubLexer>& next,
                                               const utf8_string::UTF8String& str,
                                               bool exactness)
                    {
                        // 全体的な流れは同一であり、基数の違いによる
                        // 数値の違い程度しか問題は発生しない。
                        exact_ = exactness;
                        lexer_delimiter::Normal nor;
                        utf8_string::UTF8String number = str;
                        while (true) {
                            // デリミタまで読んでしまう。
                            if (stream->isEOF())
                            {
                                throw utakata::sublexer::LexException(pos_, "illegal number format");
                            }
                        
                            utf8_string::UTF8Char ch(stream->read());
                            if (nor(ch))
                            {
                                stream->unget(ch.getBytes());
                                break;
                            }

                            number += ch;
                        }

                        // 読出したnumberを解析してNumberオブジェクトとして返す。
                        pos_ = stream->pos();
                        return lexNumber_(number);
                    }

            private:

                void setstatus_(int &now, int& prev, int next) {
                    prev = now;
                    now = next;
                }
                void setstatus_(int& now, int next) {
                    now = next;
                }

                smart_ptr<lexeme::ILexeme> lexNumber_(const utf8_string::UTF8String& number)
                    {
                        // デリミタまでを取得したデータを、実際の数値オブジェクトとして解釈させる。
                        // 解釈できなかった場合、字句構文エラーとして扱う。
                        utf8_string::UTF8String::const_utf8iterator begin = number.begin(),
                            end = number.end();

                        int status = INIT, prevstatus = INIT;
                        bool imaginary = false;
                    
                        while (begin != end) {

                            if (status == INIT)
                            {
                                // 非終端記号で言えば<complex>に当たる。
                                // 全ての整数は原則として +-0iの虚数と等しいため、そのようになる。
                                setstatus_(status, prevstatus, complex_(begin, end));
                            }
                            else if (status == REAL)
                            {
                                // 前回の状態によって、チェックが変更になる。
                                setstatus_(status, prevstatus, real_(begin, end));
                            }
                            else if (status == UREAL)
                            {
                                // 基本的にはそのまま設定するのみで構わない。ただしこの後に続く可能性があるためそのチェック
                                // も同時に行う。
                                bool b = UReal<NUM>(pos_).lex(begin, end);
                                if (b)
                                {
                                    if (!imaginary && begin != end)
                                    {
                                        // まだ続きがある。
                                        realnum_->insert(realnum_->begin(), number.begin(), begin);
                                        setstatus_(status, IMAGINARY);
                                    }
                                    else
                                    {
                                        // 上記以外の場合、基本的に終了しているはずなので返る。
                                        setstatus_(status, END);
                                        realnum_->insert(realnum_->begin(), number.begin(), begin);
                                        return lexeme::makeNumber(*realnum_, *imaginary_, exact_, NUM);
                                    }
                                }
                            }
                            else if (status == IMAGINARY)
                            {
                                // 虚数部分の判定となる。
                                setstatus_(status, prevstatus, flag_(begin, end));
                            }
                            else if (status == END) {
                                // 終了ということなので、完全に完了したことにする。
                                return lexeme::makeNumber(*realnum_, *imaginary_, exact_, NUM);
                            }
                            else if (status == FLAG)
                            {
                                // 虚数部分の先頭符号を示す。
                                utf8_string::UTF8String s;
                                s.insert(s.begin(), begin, end);

                                if (s.size() == 1)
                                {
                                    throw utakata::sublexer::LexException(pos_, "imaginary parts must constain `i' at end");
                                }
                                else if (s[s.size() - 1].toUTF16Code() != 'i')
                                {
                                    throw utakata::sublexer::LexException(pos_, "imaginary parts must constain `i' at end");
                                }

                                if (s[1].toUTF16Code() == 'i' && s.size() == 2)
                                {
                                    // 末尾がiの場合、これがそのまま文字列となる。
                                    *imaginary_ = s;
                                    return lexeme::makeNumber(*realnum_, *imaginary_, exact_, NUM);
                                }
                                else
                                {
                                    // +i,-i以外の場合は、限定領域がrealである必要がある。
                                    utf8_string::UTF8String s2;
                                    s2.insert(s2.begin(), s.begin() + 1, s.end() - 1);
                                    if (s2.toStr() == "nan.0" || s2.toStr() == "inf.0")
                                    {
                                        *imaginary_ = s;
                                        return lexeme::makeNumber(*realnum_, *imaginary_, exact_, NUM);
                                    }
                                    else
                                    {
                                        // urealである必要がある。
                                        utakata::utf8_string::UTF8String::const_utf8iterator begin = s2.begin(),
                                            end = s2.end();
                                        UReal<NUM>(pos_).lex(begin, end);
                                        // 例外が発生しなければureal_である。
                                        *imaginary_ = s;
                                        return lexeme::makeNumber(*realnum_, *imaginary_, exact_, NUM);
                                    }
                                }
                            }
                            // 各自確定するときはそのまま出力するようにしても構わない。
                            else if (status == NAN_IMAGINARY)
                            {
                                return lexeme::makeNanImaginary(number, exact_);
                            }
                            else if (status == INF_IMAGINARY)
                            {
                                return lexeme::makeInfImaginary(number, exact_);
                            }
                            else if (status == IMAGINARY_ONLY)
                            {
                                return lexeme::makeImaginaryOnly(number, exact_);
                            }
                        }
                        return smart_ptr<lexeme::ILexeme>();
                    }

                // 初期状態で確定可能なリテラルをチェックする。
                int complex_(utf8_string::UTF8String::const_utf8iterator& begin,
                             const utf8_string::UTF8String::const_utf8iterator& end)
                    {
                        // 一文字単位で調べていく。checker_に該当するデータの場合には、
                        utf8_string::UTF8String s;
                        s.insert(s.begin(), begin, end);
                        if (s.toStr() == "-nan.0i" || s.toStr() == "+nan.0i")
                        {
                            // 虚数部が正/負の非数として登録する。
                            *imaginary_ = s;
                            return NAN_IMAGINARY;
                        }
                        else if (s.toStr() == "-inf.0i" || s.toStr() == "+inf.0i")
                        {
                            // 虚数部が正/負の無限大として処理する。
                            *imaginary_ = s;
                            return INF_IMAGINARY;
                        }
                        else if (s.toStr() == "-i" || s.toStr() == "+i")
                        {
                            // 0+1i,0-1iにそれぞれマッピングされる。
                            *imaginary_ = s;
                            return IMAGINARY_ONLY;
                        }

                        // 上記以外の場合、必ず先頭はREALである必要がある。

                        return REAL;
                    }


                int real_(utf8_string::UTF8String::const_utf8iterator& begin,
                          const utf8_string::UTF8String::const_utf8iterator& end)
                    {
                        // 実数の表現でなければならない。
                        // 先頭一文字が符号表現か、符号無し表現で
                        // あること。
                        if (begin->toUTF16Code() == '+' || begin->toUTF16Code() == '-')
                        {
                            // 符号表現である場合には、以降の符号
                            // 表現までを調べる。
                            return innerReal_(begin, end, true);
                        }
                        else
                        {
                            return innerReal_(begin, end, false);
                        }
                    }

                int innerReal_(utf8_string::UTF8String::const_utf8iterator& begin,
                               const utf8_string::UTF8String::const_utf8iterator& end,
                               bool flag)
                    {
                        // realの符号有無に関係の無い、共通部分の分離。
                        if (flag)
                        {
                            ++begin;
                        }
                        utf8_string::UTF8String s;
                        utf8_string::UTF8String::const_utf8iterator save = begin;
                        while (begin != end && (begin->toUTF16Code() != '+' ||
                                                begin->toUTF16Code() != '-')) {
                            ++begin;
                        }

                        s.insert(s.begin(), save, end);
                        // とりあえず抜けた段階で、実際に抽出
                        // してみたデータを調べてみる。
                        if (s.toStr() == "nan.0" || s.toStr() == "inf.0")
                        {
                            // 非数/無限大として出力する。
                            *realnum_ = s;
                            return REAL;
                        }
                        else
                        {
                            // <uinteger>,<uinteger>/<uinteger>,<decimal 10>のいずれか
                            // である必要がある。
                            begin = save;
                            return UREAL;
                        }
                    }

            
                int flag_(utf8_string::UTF8String::const_utf8iterator& begin,
                          const utf8_string::UTF8String::const_utf8iterator& end)
                    {
                        // 虚数の先頭に必ず必要な符号が存在するかどうかを返す。
                        if (begin->toUTF16Code() == '+' || begin->toUTF16Code() == '-')
                        {
                            return FLAG;
                        }
                        // これが無い場合、特に問題は発生しない。
                        if (begin == end)
                        {
                            return END;
                        }

                        throw utakata::sublexer::LexException(pos_, "illegal number format : invalid flag previous `imaginary'");
                    }

                // チェックを行うための関数オブジェクト。
                bool exact_;
                smart_ptr<utf8_string::UTF8String> realnum_;
                smart_ptr<utf8_string::UTF8String> imaginary_;
                size_t pos_;
            };

            ////////////////////////////////
            // <uinteger>非終端記号の実装 //
            ////////////////////////////////

            template<int NUM>
            class UInteger
            {
                // 指定された基数の値であるかどうかを調べる。
                typedef utakata::utf8_string::UTF8String::const_utf8iterator iter;
            public:
                UInteger() {}
                virtual ~UInteger(){}

                bool operator()(iter it) {
                    return true;
                }
            };

            template<>
            class UInteger<2>
            {
                typedef utakata::utf8_string::UTF8String::const_utf8iterator iter;
            public:
                bool operator()(iter it) {
                    if (it->toUTF16Code() == '0' || it->toUTF16Code() == '1')
                    {
                        return true;
                    }
                    return false;
                }
            };

            template<>
            class UInteger<8>
            {
                typedef utakata::utf8_string::UTF8String::const_utf8iterator iter;
            public:
                bool operator()(iter it) {
                    if (it->toUTF16Code() == '0' || it->toUTF16Code() == '1' ||
                        it->toUTF16Code() == '2' || it->toUTF16Code() == '3' ||
                        it->toUTF16Code() == '4' || it->toUTF16Code() == '5' ||
                        it->toUTF16Code() == '6' || it->toUTF16Code() == '7')
                    {
                        return true;
                    }
                    return false;
                }
            };

            template<>
            class UInteger<10>
            {
                typedef utakata::utf8_string::UTF8String::const_utf8iterator iter;
            public:
                bool operator()(iter it) {
                    if (it->toUTF16Code() == '0' || it->toUTF16Code() == '1' ||
                        it->toUTF16Code() == '2' || it->toUTF16Code() == '3' ||
                        it->toUTF16Code() == '4' || it->toUTF16Code() == '5' ||
                        it->toUTF16Code() == '6' || it->toUTF16Code() == '7' ||
                        it->toUTF16Code() == '8' || it->toUTF16Code() == '9')
                    {
                        return true;
                    }
                    return false;
                }
            };

            template<>
            class UInteger<16>
            {
                typedef utakata::utf8_string::UTF8String::const_utf8iterator iter;
            public:
                bool operator()(iter it) {
                    if (it->toUTF16Code() == '0' || it->toUTF16Code() == '1' ||
                        it->toUTF16Code() == '2' || it->toUTF16Code() == '3' ||
                        it->toUTF16Code() == '4' || it->toUTF16Code() == '5' ||
                        it->toUTF16Code() == '6' || it->toUTF16Code() == '7' ||
                        it->toUTF16Code() == '8' || it->toUTF16Code() == '9' ||
                        it->toUTF16Code() == 'A' || it->toUTF16Code() == 'B' ||
                        it->toUTF16Code() == 'C' || it->toUTF16Code() == 'D' ||
                        it->toUTF16Code() == 'E' || it->toUTF16Code() == 'F' ||
                        it->toUTF16Code() == 'a' || it->toUTF16Code() == 'b' ||
                        it->toUTF16Code() == 'c' || it->toUTF16Code() == 'd' ||
                        it->toUTF16Code() == 'e' || it->toUTF16Code() == 'f')
                    {
                        return true;
                    }
                    return false;
                }
            };

            /////////////////////////////////
            // <ureal>非終端記号解析の実装 //
            /////////////////////////////////

            template<int NUM>
            class UReal
            {
                typedef utakata::utf8_string::UTF8String::const_utf8iterator iter;
            public:
                UReal(size_t pos) : pos_(pos)
                    {
                    }
                
                virtual ~UReal(){}


                int lex(iter& begin, const iter end)
                    {
                        // <ureal>に以降の文字列が該当するかどうかを調べる。

                        iter save = begin;

                        for (; begin != end && checker_(begin); ++begin);

                        if (begin == end)
                        {
                            // すべて数値の間に入っている場合、beginからendまでの全体が
                            // real/imaginaryのいずれかとして扱われる。
                            return true;
                        }
                        else
                        {
                            // すべてが数値の間ではなかった場合、その文字を調べる。
                            // ここに至って、10進数とそれ以外とでチェックに相違点が存在する。
                            if (begin->toUTF16Code() == '/')
                            {
                                // 次もuintegerである必要がある。
                                ++begin;
                                // マッチする間飛ばす。
                                for (; begin != end && checker_(begin); ++begin);
                                return true;
                            }
                            else
                            {
                                throw utakata::sublexer::LexException(pos_, "must have digit at least one after `/'");
                            }
                        }
                    }

            private:

                UInteger<NUM> checker_;
                size_t pos_;
            };

            template<>
            class UReal<10>
            {
                typedef utakata::utf8_string::UTF8String::const_utf8iterator iter;
            public:
                UReal(size_t pos) : pos_(pos)
                    {
                    }
                
                virtual ~UReal(){}

                bool lex(iter& begin, const iter& end)
                    {
                        // <ureal>に以降の文字列が該当するかどうかを調べる。

                        iter save = begin;
                        if (begin->toUTF16Code() == '.')
                        {
                            // 先頭が.の場合、そのまま進む。
                            for (; begin != end && checker_(begin); ++begin);
                            suffix_(begin, end);
                            return true;
                        }
                        else
                        {
                            // 先頭が.じゃない場合、一度数値を読み出しておく。
                            for (; begin != end && checker_(begin); ++begin);

                            if (begin == end)
                            {
                                //   throw utakata::sublexer::LexException(pos_, "illegal number format : not Real");
                                return true;
                            }
                            if (begin->toUTF16Code() == '.')
                            {
                                // 以降の文字がsuffixとかであるかどうかで調べる。
                                ++begin;
                                for (; begin != end && checker_(begin); ++begin);
                            }
                            suffix_(begin, end);
                            return true;
                        }
                    }

            private:

                bool suffix_(iter& begin, const iter& end)
                    {
                        // 次の文字がsuffixであるかどうかを調べる。
                        if (begin->toUTF16Code() == 'e' || begin->toUTF16Code() == 'E' ||
                            begin->toUTF16Code() == 's' || begin->toUTF16Code() == 'S' ||
                            begin->toUTF16Code() == 'f' || begin->toUTF16Code() == 'F' ||
                            begin->toUTF16Code() == 'd' || begin->toUTF16Code() == 'D' ||
                            begin->toUTF16Code() == 'l' || begin->toUTF16Code() == 'L')
                        {
                            ++begin;
                            if (begin->toUTF16Code() == '+' || begin->toUTF16Code() == '-')
                            {
                                ++begin;
                                if (checker_(begin))
                                {
                                    for (; begin != end && checker_(begin); ++begin);
                                    return true;
                                }
                                else
                                {
                                    throw utakata::sublexer::LexException(pos_, "suffix nessesary after at least one number");
                                }
                            }
                        }

                        return true;
                    }
                
                UInteger<10> checker_;
                size_t pos_;
            };
        };
    };

};

#endif /* _NUMBER_LEXER_H_ */
