#include <sstream>
#include <stack>

#include "parser.h"
#include "lexeme.h"
#include "lexer.h"
#include "lexeme_id.h"
#include "tree.h"
#include "literal.h"
#include "datum_id.h"

using namespace utakata;
using namespace utakata::interpreter;
using namespace utakata::parser;
using namespace utakata::lexeme;

parser::DatumException::DatumException(std::string str) : str_()
{
    // エラーメッセージを定義する。
    std::stringstream ss;
    ss << "datum error ! -- message : [" << str << "]" << std::endl;
    str_ = ss.str();
}

const char* parser::DatumException::what() const throw()
{
    return str_.c_str();
}

////////////////////////////
// Parser Implementations //
////////////////////////////

parser::Parser::Parser(smart_ptr<lexer::Lexer> l) : lexer_(l),
                                                     handler_()
{
    // 初期状態用ハンドラの設定
    std::pair<PARSERSTATUS, smart_ptr<IParserHandler> > p;
    p.first = PS_INIT;
    p.second.add(new InitHandler);
    handler_.insert(p);

    // リスト内部用ハンドラの設定
    p.first = PS_LIST_CAR;
    p.second.add(new ListCarHandler);
    handler_.insert(p);

    p.first = PS_LIST_CDR;
    p.second.add(new ListCdrHandler);
    handler_.insert(p);

    // .内用ハンドラの設定
    p.first = PS_AFTER_DOT;
    p.second.add(new DotHandler);
    handler_.insert(p);

    p.first = PS_END_DOT;
    p.second.add(new DotEndHandler);
    handler_.insert(p);

    // abbreviation 用ハンドラの設定
    p.first = PS_ABBREVIATION;
    p.second.add(new AbbrevHandler);
    handler_.insert(p);

    // error 用のハンドラの設定
    p.first = PS_ERROR;
    p.second.add(new ErrorHandler);
    handler_.insert(p);
}

smart_ptr<syntax::Tree> parser::Parser::parse(smart_ptr<utf8::UTF8InputStream>& strm)
{
    HandlerData data;
    data.main.add(new syntax::Tree());
    data.status.push(PS_INIT);
    data.stack.push(data.main);
    
    while(true) {

        if(data.status.top() != PS_ERROR)
        {
            data.lexm = lexer_->lex(strm);
            if(data.lexm.isNull() ||
                data.lexm->getID() == lexeme::LexemeID::eos)
            {
                break;
            }
            // 基本的にすべてハンドラに任せる。
            handler_[data.status.top()]->exec(data);
        }
        else
        {
            // ERROR だったら、強制的にハンドラを回してループから抜ける。
            handler_[data.status.top()]->exec(data);
        }
    }

    // stackのtopにはmainが殘ったままなので、data.stackのtopを追加してやる。
    
    if (data.stack.size() == 2)
    {
        data.main->push_back(data.stack.top()->begin(), data.stack.top()->end(),
                             syntax::makeCompoundDatum(syntax::DatumID::list));
    }

    return data.main;
}

bool parser::isAbbrev(smart_ptr<lexeme::ILexeme> l)
{
    // abbreviation かどうかを判別する。
    switch(l->getID().toEnum())
    {
    case lexeme::LexemeID::BACKQUOTE:        return true;
    case lexeme::LexemeID::QUOTE:            return true;
    case lexeme::LexemeID::UNQUOTE:          return true;
    case lexeme::LexemeID::UNQUOTESPLICING:  return true;
    case lexeme::LexemeID::SYNTAX:           return true;
    case lexeme::LexemeID::QUASISYNTAX:      return true;
    case lexeme::LexemeID::UNSYNTAX:         return true;
    case lexeme::LexemeID::UNSYNTAXSPLICING: return true;
    default:
        return false;
    }
}

syntax::DatumID parser::toDatumType(lexeme::LexemeID id)
{
    switch(id.toEnum())
    {
    case lexeme::LexemeID::STRING:
        return syntax::DatumID::string;
    case lexeme::LexemeID::NUMBER:
        return syntax::DatumID::number;
    case lexeme::LexemeID::IDENTIFIER:
        return syntax::DatumID::symbol;
    case lexeme::LexemeID::CHARACTOR:
        return syntax::DatumID::charactor;
    case lexeme::LexemeID::BOOLEAN:
        return syntax::DatumID::boolean;
    default:
        return syntax::DatumID::nil;
    }
}

/////////////////////
// parser handlers //
/////////////////////

bool InitHandler::exec_(HandlerData& data)
{
    // 初期状態の解釈処理を行う。
    if(data.lexm->getID() == lexeme::LexemeID::openParenthesis)
    {
        // 開き括弧である場合、新しい tree をスタックの top に設定する。
        data.stack.push(smart_ptr<syntax::Tree>(new syntax::Tree()));
        data.status.push(PS_LIST_CAR);
    }
    else if(data.lexm->getID() == lexeme::LexemeID::string ||
             data.lexm->getID() == lexeme::LexemeID::number ||
             data.lexm->getID() == lexeme::LexemeID::identifier ||
             data.lexm->getID() == lexeme::LexemeID::charactor ||
             data.lexm->getID() == lexeme::LexemeID::boolean)
    {
        // それぞれの場合、 lexeme_datum として扱われる。
        // stack の top に push_back にする。
        data.stack.top()->push_back(syntax::makeLexemeDatum(
                                  literal::generateLiteral(data.lexm),
                                  toDatumType(data.lexm->getID())));
        // status を pop する。
        data.status.pop();
    }
    else if(isAbbrev(data.lexm))
    {
        // abbreviation である場合には、 list に内包する必要があるため、
        // 新規に作成する。
        data.stack.push(smart_ptr<syntax::Tree>(new syntax::Tree()));
        data.stack.top()->push_back(syntax::makeAbbreviation(
                                  literal::generateLiteral(data.lexm)));

        // status stack に、 PS_ABBREVIATION と PS_INIT を積む。
        data.status.push(PS_ABBREVIATION);
        data.status.push(PS_INIT);
    }
    else
    {
        throw DatumException("datum の開始記号ではありません");
    }

    return true;
}

// List 内部処理のハンドリング

bool ListCarHandler::exec_(HandlerData& data)
{
    // list の開始部分である場合。
    if(data.lexm->getID() == lexeme::LexemeID::openParenthesis)
    {
        // 更に追加して積む。
        data.stack.push(smart_ptr<syntax::Tree>(new syntax::Tree()));
        // status に再度積む。
        data.status.push(PS_LIST_CAR);
    }
    else if(data.lexm->getID() == lexeme::LexemeID::closeParenthesis)
    {
        // car の時点で) が来たら、この時点で終了としておくしかない。
        smart_ptr<syntax::Tree> it = data.stack.top(); data.stack.pop();
        data.stack.top()->push_back(it->begin(), it->end(),
                                    syntax::makeCompoundDatum(syntax::DatumID::list));
        
        data.status.pop();
    }
    else if(data.lexm->getID() == lexeme::LexemeID::string ||
             data.lexm->getID() == lexeme::LexemeID::number ||
             data.lexm->getID() == lexeme::LexemeID::identifier ||
             data.lexm->getID() == lexeme::LexemeID::charactor ||
             data.lexm->getID() == lexeme::LexemeID::boolean)
    {
        //push_back して、その後に PS_LIST_CDR をセットしておく。

        data.stack.top()->push_back(syntax::makeLexemeDatum(
                                        literal::generateLiteral(data.lexm),
                                        toDatumType(data.lexm->getID())));
        data.status.pop();
        data.status.push(PS_LIST_CDR);
    }
    else if(isAbbrev(data.lexm))
    {
        // abbreviation を判別して status を返す。
        // この場合、新規にツリーを作成して、再度設定を行う。
        data.stack.push(smart_ptr<syntax::Tree>(new syntax::Tree()));
        data.stack.top()->push_back(syntax::makeAbbreviation(
                                        literal::generateLiteral(data.lexm)));

        // status stack に、 PS_ABBREVIATION と PS_INIT を積む。
        data.status.push(PS_ABBREVIATION);
        data.status.push(PS_INIT);
    }
    else
    {
        // 上記以外は許容しない。
        data.status.push(PS_ERROR);
    }

    return true;
}

bool ListCdrHandler::exec_(HandlerData& data)
{
    // list の開始部分以降である場合。
    if(data.lexm->getID() == lexeme::LexemeID::openParenthesis)
    {
        // 更に追加して積む。
        data.stack.push(smart_ptr<syntax::Tree>(new syntax::Tree()));
        // status に再度積む。
        data.status.push(PS_LIST_CAR);
    }
    else if(data.lexm->getID() == lexeme::LexemeID::closeParenthesis)
    {
        // 現在の stack の top を、下の Tree に push_back する。
        smart_ptr<syntax::Tree> it = data.stack.top(); data.stack.pop();
        data.stack.top()->push_back(it->begin(), it->end(),
                                    syntax::makeCompoundDatum(syntax::DatumID::list));
        data.status.pop();
    }
    else if(data.lexm->getID() == lexeme::LexemeID::string ||
             data.lexm->getID() == lexeme::LexemeID::number ||
             data.lexm->getID() == lexeme::LexemeID::identifier ||
             data.lexm->getID() == lexeme::LexemeID::charactor ||
             data.lexm->getID() == lexeme::LexemeID::boolean)
    {
        //push_back するが、) が来るまでは status を pop する必要は無い。
        data.stack.top()->push_back(syntax::makeLexemeDatum(
                                        literal::generateLiteral(data.lexm),
                                        toDatumType(data.lexm->getID())));
    }
    else if(isAbbrev(data.lexm))
    {
        // abbreviation を判別して status を返す。
        // この場合、新規にツリーを作成して、再度設定を行う。
        data.stack.push(smart_ptr<syntax::Tree>(new syntax::Tree()));
        data.stack.top()->push_back(syntax::makeAbbreviation(
                                        literal::generateLiteral(data.lexm)));

        // status stack に、 PS_ABBREVIATION と PS_INIT を積む。
        data.status.push(PS_ABBREVIATION);
    }
    else if(data.lexm->getID() == lexeme::LexemeID::dot)
    {
        // .が来た場合、これは許容される。
        // この場合、.が完了した時点でリストが終了していなければならないため、
        // after_dot の後には、 end_list ハンドラが設定される。
        data.status.push(PS_AFTER_DOT);
    }
    else
    {
        // 上記以外は許容しない。
        data.status.push(PS_ERROR);
    }

    return true;
}

bool DotEndHandler::exec_(HandlerData& data)
{
    // .後の) を待つ部分のハンドラ。
    // このハンドラでは、) 以外が来た場合にはすべて構文エラーとなる。
    if(data.lexm->getID() == lexeme::LexemeID::closeParenthesis)
    {
        // 現在の stack の top を、下の Tree に push_back する。
        smart_ptr<syntax::Tree> it = data.stack.top(); data.stack.pop();
        data.stack.top()->push_back(it->begin(), it->end(),
                                    syntax::makeCompoundDatum(syntax::DatumID::list));
        data.status.pop();
    }
    else
    {
        // 上記以外は許容しない。
        data.status.push(PS_ERROR);
    }

    return true;
}

// dot 処理内部のハンドリング

bool DotHandler::exec_(HandlerData& data)
{
    // .の後である場合。
    data.status.pop();
    if(data.lexm->getID() == lexeme::LexemeID::openParenthesis)
    {
        // 更に追加して積む。
        data.stack.push(smart_ptr<syntax::Tree>(new syntax::Tree()));
        // status に再度積む。
        data.status.push(PS_END_DOT);
        data.status.push(PS_LIST_CAR);
    }
    else if(data.lexm->getID() == lexeme::LexemeID::string ||
             data.lexm->getID() == lexeme::LexemeID::number ||
             data.lexm->getID() == lexeme::LexemeID::identifier ||
             data.lexm->getID() == lexeme::LexemeID::charactor ||
             data.lexm->getID() == lexeme::LexemeID::boolean)
    {
        // push_back した後、この時点で after_dot が完了しているため、
        // end_dot に移動する。

        data.stack.top()->push_back(syntax::makeLexemeDatum(
                                        literal::generateLiteral(data.lexm),
                                        toDatumType(data.lexm->getID())));
        data.status.push(PS_END_DOT);
    }
    else if(isAbbrev(data.lexm))
    {
        // abbreviation を判別して status を返す。
        // この場合、新規にツリーを作成して、再度設定を行う。
        data.stack.push(smart_ptr<syntax::Tree>(new syntax::Tree()));
        data.stack.top()->push_back(syntax::makeAbbreviation(
                                        literal::generateLiteral(data.lexm)));

        // status stack に、 PS_END_DOT まで含めて、
        // PS_ABBREVIATION と PS_INIT を積む。
        data.status.push(PS_END_DOT);
        data.status.push(PS_ABBREVIATION);
        data.status.push(PS_INIT);
    }
    else
    {
        // 上記以外はすべて不正となる。
        data.status.push(PS_ERROR);
    }

    return true;
}

// ABBREVIATION 処理のハンドリング。

bool AbbrevHandler::exec_(HandlerData& data)
{
    // この場合、 top の一つ下が、 abbrev が属する list の親になっているため、
    // この場合でも単純に list として設定できる。
    if(data.lexm->getID() == lexeme::LexemeID::openParenthesis)
    {
        // 更に追加して積む。
        data.stack.push(smart_ptr<syntax::Tree>(new syntax::Tree()));
        // status に再度積む。
        data.status.push(PS_LIST_CAR);
    }
    else if(data.lexm->getID() == lexeme::LexemeID::string ||
            data.lexm->getID() == lexeme::LexemeID::number ||
            data.lexm->getID() == lexeme::LexemeID::identifier ||
            data.lexm->getID() == lexeme::LexemeID::charactor ||
            data.lexm->getID() == lexeme::LexemeID::boolean)
    {
        //push_back するが、) が来るまでは status を pop する必要は無い。
        data.stack.top()->push_back(syntax::makeLexemeDatum(
                                        literal::generateLiteral(data.lexm),
                                        toDatumType(data.lexm->getID())));
    }
    else if(isAbbrev(data.lexm))
    {
        // abbreviation を判別して status を返す。
        // この場合、新規にツリーを作成して、再度設定を行う。
        data.stack.push(smart_ptr<syntax::Tree>(new syntax::Tree()));
        data.stack.top()->push_back(syntax::makeAbbreviation(
                                        literal::generateLiteral(data.lexm)));

        // status stack に、 PS_ABBREVIATION と PS_INIT を積む。
        data.status.push(PS_ABBREVIATION);
    }
    else if(data.lexm->getID() == lexeme::LexemeID::dot)
    {
        // .が来た場合、これは許容される。
        // この場合、.が完了した時点でリストが終了していなければならないため、
        // after_dot の後には、 end_list ハンドラが設定される。
        data.status.push(PS_AFTER_DOT);
    }
    else
    {
        // 上記以外は許容しない。
        data.status.push(PS_ERROR);
    }

    
    if (data.stack.top()->size() == 3)
    {
        smart_ptr<syntax::Tree> it = data.stack.top(); data.stack.pop();
        data.stack.top()->push_back(it->begin(), it->end(),
                                    syntax::makeCompoundDatum(syntax::DatumID::list));

        // 最後に pop しておけばそれで問題無い。
        data.status.pop();
    }

    return true;
}

// syntax error のハンドリング

bool ErrorHandler::exec_(HandlerData& data)
{
    // 基本的にはエラーのハンドリングを行うだけ。
    // 一応発生した前後の位置と発生した lexeme くらいは表示したいかな。
    std::stringstream ss;
    ss << "syntax error" << data.stack.top()->size() << data.stack.top()->begin()->second->toValue()->toStr();
    throw DatumException(ss.str());

    return true;
}
