#ifndef _PARSER_H_
#define _PARSER_H_

#include <string>
#include <vector>
#include <map>

#include "smart_ptr.h"
#include "parser_handler.h"
#include "parser_status.h"

namespace utakata {

    namespace syntax {
        class Tree;
        class DatumID;
    };

    namespace lexer {
        class Lexer;
    };

    namespace utf8 {
        class UTF8InputStream;
    };

    namespace literal {
        class Literal;
    };

    namespace parser {
    
        class DatumException : public std::exception
        {
        public:
            DatumException(std::string str);
            virtual ~DatumException() throw() {}
            const char* what() const throw();

        private:

            std::string str_;
        };

        class Parser
        {
            // lexerを受け取り、lexerが保持する全体の解析を行いつつ、
            // 実行を行う。
            // 実行は、基本的に全体を読み込み、全体を保持した状態で行う。
            
        public:
            
            Parser(smart_ptr<lexer::Lexer> l);
            virtual ~Parser() {}

            // 受け取ったlexerからの字句の取得を行い、解析を行なう。
            // 成功した場合には、先頭から順番にdatumとして読み込んだObjectの
            // vectorを返す。
            // 致命的なエラーが発生した場合には例外を返す。
            smart_ptr<syntax::Tree> parse(smart_ptr<utf8::UTF8InputStream>& strm);

        private:
            smart_ptr<lexer::Lexer> lexer_;
            // 各PARSERSTATUSに関連づけられたParserHandlerのmap
            std::map<PARSERSTATUS, smart_ptr<IParserHandler> > handler_;
        };

        // 渡したlexemeがAbbrevationであるかどうかを調べる。
        bool isAbbrev(smart_ptr<lexeme::ILexeme> l);

        syntax::DatumID toDatumType(lexeme::LexemeID id);

        // 各ハンドラの定義。

        class InitHandler : public IParserHandler
        {
            // PS_INITに対応するハンドラ
        public:
            InitHandler() {}
            virtual ~InitHandler() {}

        private:

            bool exec_(HandlerData &data);

        };

        // <list>に対応するものだけは、carとcdrでハンドラが分離している。
        class ListCarHandler : public IParserHandler
        {
            // PS_LIST_CARである間に対応するハンドラ
        public:
            ListCarHandler() {}
            virtual ~ListCarHandler() {}

        private:

            bool exec_(HandlerData& data);

        };

        
        class ListCdrHandler : public IParserHandler
        {
            // PS_LIST_CDRである間に対応するハンドラ
        public:
            ListCdrHandler() {}
            virtual ~ListCdrHandler() {}

        private:

            bool exec_(HandlerData& data);

        };


        class DotHandler : public IParserHandler
        {
            // PS_AFTER_DOTに対応するハンドラ
        public:
            DotHandler() {}
            virtual ~DotHandler() {}

        private:

            bool exec_(HandlerData& data);

        };

        // <list>に対応するものだけは、carとcdrでハンドラが分離している。
        class DotEndHandler : public IParserHandler
        {
            // PS_END_DOTに対応するハンドラ。
        public:
            DotEndHandler() {}
            virtual ~DotEndHandler() {}

        private:

            bool exec_(HandlerData& data);

        };


        class AbbrevHandler : public IParserHandler
        {
            // PS_ABBREVIATIONに対応するハンドラ
        public:
            AbbrevHandler() {}
            virtual ~AbbrevHandler() {}

        private:

            bool exec_(HandlerData& data);

        };

        class ErrorHandler : public IParserHandler
        {
            // PS_ERRORをハンドリングする。実際にはエラーを通告する
            // のみ。
        public:
            ErrorHandler() {}
            virtual ~ErrorHandler() {}
            
        private:

            bool exec_(HandlerData& data);

        };

    };

};

#endif /* _PARSER_h_ */
