#ifndef ___HGL___REF_DELETE___
#define ___HGL___REF_DELETE___

/**
 ref_delete
 参照カウントを保持して、
 0になったときに、保持してあるオブジェクトを破棄するための
 クラス。
 smart_ptr中で、配列と非配列オブジェクトとを区別するために
 使用される。
*/
class ref_delete_base {
public:

    virtual ~ref_delete_base() {}

    /**
     dec_ref
     参照カウントのデクリメント。
     これを実行したとき、参照カウントが0になったら、do_deleteを
     実行する。
     @access public
     @return bool true:削除済み false:未削除
    */
    bool dec_ref() {
        if (--m_nRef <= 0) {
            if (m_bOwner) {
                do_delete();
            }
            return true;
        }
        return false;
    }

    /**
     inc_ref
     参照カウントのインクリメント
     @access public
     @return void
    */
    void inc_ref() {
        m_nRef++;
    }

    /**
     getRef
     参照カウントの取得
     @access public
     @return int 参照カウント
    */
    int getRef() const {return m_nRef;}
    
    /**
     setRef
     参照カウントの設定
     @param int i 参照カウント
     @access public
     @return void
    */
    void setRef(int i) {m_nRef = i;}

    /**
     setOwner
     所有権の設定
     @param bool b 所有権の有無
     @access public
     @return void
    */
    void setOwner(bool b) {m_bOwner = b;}
    
    /**
     getOwner
     所有権の取得
     @access public
     @return bool
    */
    bool getOwner() {return m_bOwner;}

    /**
     getObjSize
     オブジェクトサイズの取得
     @access public
     @return int オブジェクトサイズ
    */
    int getObjSize() const {return m_nObjSize;}
    
    /**
     setObjSize
     オブジェクトサイズの設定
     @param int i オブジェクトサイズ
     @access public
     @return void
    */
    void setObjSize(int i) {m_nObjSize = i;}

    /**
     getMaxNum
     最大インデックスの取得
     @access public
     @return int 最大インデックス
    */
    int getMaxNum() const {return m_nMaxNum;}
    
    /**
     setMaxNum
     最大インデックスの設定
     @param int i 最大インデックス
     @access public
     @return void
    */
    void setMaxNum(int i) {m_nMaxNum = i;}

    /**
     do_delete
     参照カウントがゼロになった際に実行される。
     主に、deleteとdelete[]の違いを埋めるために使う
     @access protected
     @return void
    */
    virtual void do_delete() = 0;

protected:

    int m_nRef;         //参照カウント
    bool m_bOwner;      //所有権
    int m_nMaxNum;      //最大インデックス
    int m_nObjSize;     //オブジェクトのサイズ

};

/**
 ref_noarray_object
 非配列オブジェクトに対する、
 参照カウントを保持した削除クラス
*/
template<class T>
class ref_noarray_object : public ref_delete_base {
public:

    ref_noarray_object(T* p) : m_pObj(p) {}
    
    virtual void do_delete() {
        if (m_pObj != NULL) {
            delete m_pObj;
        }
    }

protected:

    T* m_pObj;

};

/**
 ref_array_object
 配列オブジェクトに対する、
 参照カウントを保持した削除クラス
*/
template<class T>
class ref_array_object : public ref_delete_base {
public:

    ref_array_object(T* p) : m_pObj(p) {}
    
    virtual void do_delete() {

        if (m_pObj != NULL) {
            delete [] m_pObj;
        }
    }

protected:

    T* m_pObj;

};

/**
 ref_null_object
 所有権を持たないオブジェクトに対する、
 参照カウントを保持した削除クラス
*/
template<class T>
class ref_null_object : public ref_delete_base{
public:

    ref_null_object(T* p) : m_pObj(p) {}
    
    virtual void do_delete() {
    }

protected:

    T* m_pObj;

};

#endif
