#include <vector>
#include <iostream>
#include <string>
#include <algorithm>

#include "simpletest.h"

using namespace simpletest;

SimpleTestSuite::SimpleTestSuite(const std::string& name) :
    name_("[" + name + "]"), prefix_(">>"), testers_(), strm_(std::cout),
    runner_()
{
    runner_.add(new SimpleTestRunner(strm_, prefix_));
}

SimpleTestSuite::SimpleTestSuite(const std::string& name,
                                   std::ostream& strm) :
    name_("[" + name + "]"), prefix_(">>"), testers_(), strm_(strm)
{
    runner_.add(new SimpleTestRunner(strm_, prefix_));
}

bool SimpleTestSuite::addTester(const sfcr::sfc_r<bool>::type& func)
{
    // テストを追加する。中身がNULLの場合には触らない。
    if (func.isNull())
    {
        return false;
    }
    testers_.push_back(func);
    return true;
}

void SimpleTestSuite::run()
{
    // この場合、標準出力される、std::coutが利用される。
    // 現在保持しているテストをまとめて実行する。
    // 実行した結果は、全て出力されるようになる。
    
    printBeginMessage();

    SimpleTestRunner runner(strm_, prefix_);
    std::for_each(testers_.begin(), testers_.end(),
                  *runner_);

    printEndMessage();
}

void simpletest::SimpleTestSuite::printBeginMessage()
{
    // 開始メッセージを、指定されたストリームに出力する。

    const std::string separates("==================================================");
    const std::string run(" is running...");
    const std::string suite("Test suite is ");

    strm_ << separates << std::endl;
    strm_ << prefix_ << suite << name_ << run << std::endl;
    strm_ << separates << std::endl;
}

void simpletest::SimpleTestSuite::printEndMessage()
{
    // テストの終了メッセージを、指定されたストリームに出力する。
    const std::string separates("==================================================");
    const std::string message("SimpleTest Results.");
    
    strm_ << separates << std::endl;
    strm_ << prefix_ << message << std::endl;

    // 結果を利用して出力する。
    {
        const std::string mes("executes tests ");
        int count = testers_.size();
 
        strm_ << prefix_ << mes << count << " --> ";
        strm_ << "success [" << runner_->getSuccess() << "] : ";
        strm_ << "fail [" << runner_->getFail() << "] : ";
        strm_ << "exception [" << runner_->getException() << "]" << std::endl;
    }

    strm_ << separates << std::endl;
}

smart_ptr<SimpleTestAsserter> simpletest::SimpleTestSuite::getAsserter()
{
    return smart_ptr<simpletest::SimpleTestAsserter>(new
                                                      simpletest::SimpleTestAsserter(strm_, prefix_));
}


simpletest::SimpleTestRunner::SimpleTestRunner(std::ostream& strm,
                                                 const std::string& prefix) :
    strm_(strm), prefix_(prefix), pImpl(new simpletest::SimpleTestRunner::PImpl)
{}

void simpletest::SimpleTestRunner::innerCheck(bool b)
{
    // 関数が例外を返さなかった場合に実行されるチェック関数
    pImpl->count++;
    if (b)
    {
        pImpl->success++;
        printSuccess();
    } else {
        pImpl->fail++;
        printFail();
    }
}

void simpletest::SimpleTestRunner::innerCheck(std::exception& e) {
    // 例外が発行された場合、とりあえずはe.whatsを
    // std::stringとして保持しておく。
    pImpl->exception++;
    printException(e);
}

void simpletest::SimpleTestRunner::printSuccess()
{
    // 成功した場合のメッセージを出力する。
    int count = pImpl->success + pImpl->fail + pImpl->exception;
    strm_ << prefix_ << "test number ";
    strm_ << count << " is success." << std::endl;
}

void simpletest::SimpleTestRunner::printFail()
{
    // 成功した場合のメッセージを出力する。
    int count = pImpl->success + pImpl->fail + pImpl->exception;
    strm_ << prefix_ << "test number ";
    strm_ << count << " is fail." << std::endl;
}

void simpletest::SimpleTestRunner::printException(const std::exception& e)
{
    // 成功した場合のメッセージを出力する。
    int count = pImpl->success + pImpl->fail + pImpl->exception;
    strm_ << prefix_ << "test number ";
    strm_ << count << " is throw exception." << std::endl;
    strm_ << e.what() << std::endl;
}


//================================================================================


simpletest::SimpleTestAsserter::SimpleTestAsserter(std::ostream& strm,
                                                     const std::string& prefix) :
    strm_(strm), prefix_(prefix), count_(0)
{
}

simpletest::SimpleTestAsserter::SimpleTestAsserter(const SimpleTestAsserter& asserter) :
    strm_(asserter.strm_), prefix_(asserter.prefix_), count_(asserter.count_)
{
}

SimpleTestAsserter& simpletest::SimpleTestAsserter::operator=(const SimpleTestAsserter& asserter)
{
    // 代入によってコピーを作成するための処理
    SimpleTestAsserter a(asserter);
    std::swap(*this, a);
    return *this;
}

bool simpletest::SimpleTestAsserter::isOk() const
{
    // count_は、checkの比較関数が失敗する度にこれがインクリメントされることに
    // なっている。
    return count_ == 0 ? true : false;
}
