#ifndef _STACK_H_
#define _STACK_H_

#include <exception>

namespace vm {

    class stack_overflow : public std::exception
    {
    public:
        const char* what() const throw()
            {
                return "stack overflow";
            }
    };

    /**
       VM内部で汎用的に利用可能なスタックを定義する。
    */
    template<class Ty, int Size>
    class stack {
    public:

        stack() :  size_(0) {}
        virtual ~stack() {
            resize(0);
        }

        /**
           スタックにデータを一つ追加する。
        */
        void push(const Ty& value) {
            if (Size <= size_) {
                throw stack_overflow();
            }
            *(::new(this->data_[this->size_++]) Ty) = value;
        }
        
        /**
           スタックからデータを一つ取りだす。
        */
        void pop() {
            // 一度Tyのポインタとしてキャストし、先にデクリメントしたsize_の位置のデータ
            // のデストラクタを直接呼びだす。
            reinterpret_cast<Ty*>(this->data_[--(this->size_)])->~Ty();
        }

        void pop(int count) {
            resize(size_ - count);
        }

        /**
           指定したスタックサイズにする。
        */
        void resize(int newsize) {
            int oldsize = this->size_;

            // 前のスタックサイズよりも大きかった場合
            if (oldsize > newsize) {
                // 既に存在しているのから開放していくため、オーバーフローは発生しない。
                for (int i = newsize; i < oldsize; ++i)
                {
                    reinterpret_cast<Ty *>(this->data_[i])->~Ty();
                }
            }

            // 前のスタックサイズよりも小さかった場合
            if (oldsize < newsize)
            {
                // スタックのサイズよりも大きかった場合には、例外を発生させる。
                if (Size < newsize) {
                    throw stack_overflow();
                }

                // 差分の分だけ作成する。
                for (int i = oldsize; i < newsize; ++i)
                {
                    // placement new
                    // ここは、newという関数を直接呼びだしている。newは指定された領域から、
                    // 指定されたサイズ分だけの領域を作成して、割り当てたメモリを返す。
                    ::new(this->data_[i]) Ty;
                }
            }
            this->size_ = newsize;
        }

        /**
           スタックのトップにあるオブジェクトを取得する。
        */
        const Ty& top() const {return *reinterpret_cast<const Ty*>(this->data_[this->size_ - 1]);}
        Ty& top() {return *reinterpret_cast<Ty*>(this->data_[this->size_ - 1]);}

        bool overflow() const {return this->size_ >= Size;}
        bool empty() const {return this->size_ == 0;}
        int size() const {return this->size_;}

        /**
           スタックのインデックスを指定し、指定した位置のスタックを取得する。
        */
        const Ty& operator[](int index) const {
            return *reinterpret_cast<const Ty *>(this->data_[index]);}
        Ty& operator[](int index) {
            return *reinterpret_cast<Ty *>(this->data_[index]);}


    private:

        // 各データを積むためのデータ
        char data_[Size][sizeof(Ty)];

        // 現在のスタックインデックス
        int size_;
    };
};

#endif /* _STACK_H_ */
