#ifndef _SUBLEXER_IMPL_H_
#define _SUBLEXER_IMPL_H_

#include <string>

#include "sublexer.h"
#include "smart_ptr.h"
#include "utf8.h"
#include "lexeme.h"
#include "utf8_string.h"


namespace utakata {

    namespace sublexer {

        class LexException : public std::exception
        {
        public:
            LexException(size_t pos, std::string str);
            virtual ~LexException() throw() {}
            const char* what() const throw();

        private:

            size_t pos_;
            std::string str_;
        };

        class FirstLexer : public ISubLexer
        {
            // lex関数の内部で最初に実行される。
            // ここから返されない場合は、一文字の構文構成文字が返されるということになる。
        public:
            FirstLexer() {}
            virtual ~FirstLexer() {}

            smart_ptr<lexeme::ILexeme> lex(smart_ptr<utakata::utf8::UTF8InputStream> stream,
                                           smart_ptr<ISubLexer>& next);

        private:

            // lex関数中で利用される、雑多になった字句解析関数の分割
            smart_ptr<lexeme::ILexeme> lex_(const utakata::utf8_string::UTF8String& str,
                                            smart_ptr<utakata::utf8::UTF8InputStream> stream,
                                            smart_ptr<ISubLexer>& next);

        };

        class StringLexer : public ISubLexer
        {
            // 文字列だと判断された場合に実行されるsublexer。
            // 文字列が不正に終わったりしている場合には、その時点で例
            // 外を発行する。
        public:
            StringLexer() {}
            virtual ~StringLexer() {}

            smart_ptr<lexeme::ILexeme> lex(smart_ptr<utakata::utf8::UTF8InputStream> stream,
                                           smart_ptr<ISubLexer>& next);
        };

        class OneLineCommentLexer : public ISubLexer
        {
            // コメントを返す。
            // と言っても、;から開始されるコメントは完全に無視されるので、lexeme
            // を保存する必要すらない。
        public:

            OneLineCommentLexer(){}
            virtual ~OneLineCommentLexer(){}

            smart_ptr<lexeme::ILexeme> lex(smart_ptr<utakata::utf8::UTF8InputStream> stream,
                                           smart_ptr<ISubLexer>& next);
        };

        class NestedCommentLexer : public ISubLexer
        {
            // コメントを返す。
            // と言っても、;から開始されるコメントは完全に無視されるので、lexeme
            // を保存する必要すらない。
        public:

            NestedCommentLexer(){}
            virtual ~NestedCommentLexer(){}

            smart_ptr<lexeme::ILexeme> lex(smart_ptr<utakata::utf8::UTF8InputStream> stream,
                                           smart_ptr<ISubLexer>& next);
        };

        class CharactorLexer : public ISubLexer
        {
            // 文字名、及び16進数による文字指定の解析を行なう。
            // #\文字名、#\x<hex>;の両方を解釈する。
        public:

            CharactorLexer(){}
            virtual ~CharactorLexer(){}

            smart_ptr<lexeme::ILexeme> lex(smart_ptr<utakata::utf8::UTF8InputStream> stream,
                                           smart_ptr<ISubLexer>& next);

        };

        class NumberLexer : public ISubLexer
        {
            // 数値を解析する。
            // ここで解析する数値は、<number>に相当する。
            const unsigned char BINARY;
            const unsigned char OCTET;
            const unsigned char DECIMAL;
            const unsigned char HEX;
        public:

            NumberLexer(const utakata::utf8_string::UTF8String& str);
            virtual ~NumberLexer(){}

            smart_ptr<lexeme::ILexeme> lex(smart_ptr<utakata::utf8::UTF8InputStream> stream,
                                           smart_ptr<ISubLexer>& next);
        private:

            smart_ptr<lexeme::ILexeme> innerLex_(smart_ptr<utakata::utf8::UTF8InputStream> stream,
                                                 smart_ptr<ISubLexer>& next,
                                                 const utakata::utf8_string::UTF8String& str);


            // 渡した文字列からprefixを抽出して返す。
            unsigned char getPrefix_(const utakata::utf8_string::UTF8String& str);
            // 正確性を抽出して返す。
            void checkExactness_(const utakata::utf8_string::UTF8String& str); 

        private:
            
            smart_ptr<utakata::utf8_string::UTF8String> str_;

            bool exact_;        // 正確性を設定する。
            unsigned char prefix_;       // prefixを設定する。
            
        };

        class IdentifierLexer : public ISubLexer
        {
            // コメントを返す。
            // と言っても、;から開始されるコメントは完全に無視されるので、lexeme
            // を保存する必要すらない。
        public:

            IdentifierLexer(const utakata::utf8_string::UTF8String& str);
            virtual ~IdentifierLexer(){}

            smart_ptr<lexeme::ILexeme> lex(smart_ptr<utakata::utf8::UTF8InputStream> stream,
                                           smart_ptr<ISubLexer>& next);

        private:

            smart_ptr<utakata::utf8_string::UTF8String> str_;
        };

        class BooleanLexer : public ISubLexer
        {
            // #t、または#fを返す。
            // いずれかの次には、必ず区切り文字がなければならない。
        public:

            BooleanLexer(const utakata::utf8_string::UTF8Char& ch);
            virtual ~BooleanLexer(){}

            smart_ptr<lexeme::ILexeme> lex(smart_ptr<utakata::utf8::UTF8InputStream> stream,
                                           smart_ptr<ISubLexer>& next);

        private:

            smart_ptr<utakata::utf8_string::UTF8Char> ch_;
        };

        class ByteVectorLexer : public ISubLexer
        {
            // #vu8であるかどうかを検査する。
            // vu8ではない場合には、エラーとなる。
        public:

            ByteVectorLexer(const utakata::utf8_string::UTF8Char& ch);
            virtual ~ByteVectorLexer(){}

            smart_ptr<lexeme::ILexeme> lex(smart_ptr<utakata::utf8::UTF8InputStream> stream,
                                           smart_ptr<ISubLexer>& next);

        private:

            smart_ptr<utakata::utf8_string::UTF8Char> ch_;
        };

    };

};


#endif /* _SUBLEXER_IMPL_H_ */
