#include <iostream>
#include <sstream>
#include <string>
#include <functional>

#include "../simpletest.h"

#include "../textarrayformat.h"
#include "../lexeme_id.h"
#include "../utf8.h"
#include "../lexer.h"
#include "../lexeme.h"

using namespace std;
using namespace utakata;

bool lexer_test(smart_ptr<simpletest::SimpleTestAsserter> asserter)
{
    std::stringstream ss;
    ss << "==========" << endl;
    ss << "(hoge 12 \"hoge\" )" << endl;

    textarrayformat::TextArrayReader reader(ss);
    smart_ptr<istream> formats(new stringstream(reader.get()));
    smart_ptr<utakata::utf8::UTF8InputStream> st;
    st.add(new utakata::utf8::UTF8InputStream(formats));
    
    utakata::lexer::Lexer lexer;

    smart_ptr<lexeme::ILexeme> m(lexer.lex(st));
    asserter->checkOnly(m->toString()->toStr(), "(");
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::openParenthesis.toEnum(), "open paren");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::identifier.toEnum(), "identifier");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::number.toEnum(), "number");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::string.toEnum(), "string");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::closeParenthesis.toEnum(), "close");

    // ここで終了しているはず。EOFの状態から実行されると、EOFのIDを持った
    // lexemeが返される。
    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::eos.toEnum(), "eos");
    
    return asserter->isOk();
}

bool lexer_test2(smart_ptr<simpletest::SimpleTestAsserter> asserter)
{
    std::stringstream ss;
    ss << "==========" << endl;
    ss << ",@(hoge 12 `(hoge) 'huga #f #t)" << endl;

    textarrayformat::TextArrayReader reader(ss);
    smart_ptr<istream> formats(new stringstream(reader.get()));
    smart_ptr<utakata::utf8::UTF8InputStream> st;
    st.add(new utakata::utf8::UTF8InputStream(formats));
    
    utakata::lexer::Lexer lexer;

    smart_ptr<lexeme::ILexeme> m(lexer.lex(st));
    asserter->checkOnly(m->toString()->toStr(), ",@");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::openParenthesis.toEnum(), "open paren");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::identifier.toEnum(), "identifier");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::number.toEnum(), "number");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::backquote.toEnum(), "back quote");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::openParenthesis.toEnum(), "open paren2");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::identifier.toEnum(), "identifier2");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::closeParenthesis.toEnum(), "close");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::quote.toEnum(), "quote");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::identifier.toEnum(), "identifier2");

    // booleanの解析が行えるかどうか。
    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::boolean.toEnum(), "boolean:false");
    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::boolean.toEnum(), "boolean:true");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::closeParenthesis.toEnum(), "close");

    // ここで終了しているはず。EOFの状態から実行されると、EOFのIDを持った
    // lexemeが返される。
    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::eos.toEnum(), "eos");
    
    return asserter->isOk();
}

bool lexer_test3(smart_ptr<simpletest::SimpleTestAsserter> asserter)
{
    std::stringstream ss;
    ss << "==========" << endl;
    ss << "#'(hoge #,12 #`(hoge) 'huga) ; comment" << endl;
    ss << "hoge #vu8(12 12) #(vector)" << endl;

    textarrayformat::TextArrayReader reader(ss);
    smart_ptr<istream> formats(new stringstream(reader.get()));
    smart_ptr<utakata::utf8::UTF8InputStream> st;
    st.add(new utakata::utf8::UTF8InputStream(formats));
    
    utakata::lexer::Lexer lexer;

    smart_ptr<lexeme::ILexeme> m(lexer.lex(st));
    asserter->checkOnly(m->toString()->toStr(), "#'", "syntax");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::openParenthesis.toEnum(), "open paren");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::identifier.toEnum(), "identifier");

    m = lexer.lex(st);
    asserter->checkOnly(m->toString()->toStr(), "#,", "unsyntax");
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::unsyntax.toEnum(), "unsyntax");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::number.toEnum(), "number");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::quasiSyntax.toEnum(), "quasi syntax");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::openParenthesis.toEnum(), "open paren2");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::identifier.toEnum(), "identifier2");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::closeParenthesis.toEnum(), "close");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::quote.toEnum(), "quote");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::identifier.toEnum(), "identifier2");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::closeParenthesis.toEnum(), "close");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::identifier.toEnum(), "identifier3");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::byteVector.toEnum(), "byteVector");
    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::number.toEnum(), "byteVector:number1");
    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::number.toEnum(), "byteVector:number2");
    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::closeParenthesis.toEnum(), "close");

    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::vector.toEnum(), "vector");
    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::identifier.toEnum(), "vector:identifier");
    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::closeParenthesis.toEnum(), "close");
    
    // ここで終了しているはず。EOFの状態から実行されると、EOFのIDを持った
    // lexemeが返される。
    m = lexer.lex(st);
    asserter->checkOnly(m->getID().toEnum(), lexeme::LexemeID::eos.toEnum(), "eos");
    
    return asserter->isOk();
}


int main(int argc, char *argv[])
{
    simpletest::SimpleTestSuite suite("main lexer test");
    suite.addTester(sfcr::screate(lexer_test, suite.getAsserter()));
    suite.addTester(sfcr::screate(lexer_test2, suite.getAsserter()));
    suite.addTester(sfcr::screate(lexer_test3, suite.getAsserter()));
    suite.run();
    return 0;
}
