#include <ios>
#include <iostream>
#include <sstream>
#include <string>
#include <functional>


#include "../simpletest.h"

#include "../utf8.h"

#include "../utf8_string.h"
#include "../textarrayformat.h"

bool utf8_multichar_test(smart_ptr<simpletest::SimpleTestAsserter> asserter)
{

    std::string tmp("あいuえお");
    // マルチバイト文字列を正しく読みだせるかどうかのチェックを
    // 含めたテスト
    smart_ptr<std::istream> ss(new std::stringstream(tmp));

    utakata::utf8::UTF8InputStream stream(ss);

    // 単独のreadを試す。
    utakata::utf8_string::UTF8Char ch(stream.read());
    asserter->check(ch.toUTF16Code(), utakata::utf8::generateUTF8Code("あ"));
    // peekが問題なく働いていることのテスト
    utakata::utf8_string::UTF8Char ch2(stream.peek());
    asserter->check(ch2.toUTF16Code(), utakata::utf8::generateUTF8Code("い"));
    utakata::utf8_string::UTF8Char ch3(stream.read());
    asserter->check(ch3.toUTF16Code(), utakata::utf8::generateUTF8Code("い"));

    // asciiだとしても問題無く読みだせるはず。
    utakata::utf8_string::UTF8Char ch4(stream.peek());
    asserter->check(ch4.toUTF16Code(), 'u');

    // ascii文字の範囲であるかどうかを判定する関数のテスト。
    asserter->check(utakata::utf8_string::is_ascii_char(ch4), true);
    asserter->check(utakata::utf8_string::is_ascii_char(ch), false);

    // 同一コードなので比較して同じになるはず。
    asserter->check(ch == ch2, false, "chとch2が異なる");
    asserter->check(ch < ch2, true , "chよりch2が大きいはず");

    // 文字として直接返せる。ただし比較はstd::stringでしか行えない。
    asserter->check(ch2.toStr(), "い");

    return asserter->isOk();
}

bool utf8_string_test(smart_ptr<simpletest::SimpleTestAsserter> asserter)
{
    // マルチバイトとascii文字の混在文字も正しく扱うことのできる
    // CUTF8Stringのテスト
    
    std::string tmp("あいうえsssお(");

    smart_ptr<std::istream> ss(new std::stringstream(tmp));
    utakata::utf8::UTF8InputStream stream(ss);

    utakata::utf8_string::UTF8String str(stream.read(5));
    asserter->check(str.begin()->toStr(), "あ");
    asserter->check(str.size(), 5);

    // 自分自身との加算でも問題ないことを確認する。
    str += str;
    asserter->check(str[5].toStr(), "あ");

    // 本当に基本的なもの以外は、同一の名前空間内に関数が用意されている。
    // utf8_stringから返されるのは、UTF8のバイト列を設定した文字列であると
    // する。
    asserter->check(utakata::utf8_string::substring(str,0,2).toStr(), "あい");
    asserter->check(utakata::utf8_string::substring(str,1,2).toStr(), "いう");
    asserter->check(utakata::utf8_string::substring(str, 2).toStr(), "うえsあいうえs");

    {
        ss->seekg(0, std::ios::beg);
        utakata::utf8_string::UTF8String str(stream.read(9));
        str += std::string("(");
        asserter->check(str.toStr(), "あいうえsssお((");
    }
    // 文字列とするのは自身の関数だけで可能。
    asserter->check(str.toStr(), "あいうえsあいうえs");

    return asserter->isOk();
}

bool utf8_string_util_test(smart_ptr<simpletest::SimpleTestAsserter> asserter)
{
    // CUTF8Stringと共に利用するためのユーティリティ関数の操作を
    // 行う。
    
    std::string tmp("あいうえsssお");

    smart_ptr<std::istream> ss(new std::stringstream(tmp));
    utakata::utf8::UTF8InputStream stream(ss);

    utakata::utf8_string::UTF8String str(stream.read(5));
    utakata::utf8_string::UTF8String str2(stream.read(3));
    utakata::utf8_string::UTF8String str3 = str;

    // 挿入してみる。
    str.insert(str.begin(), str2.begin(), str2.end());
    asserter->check(str.toStr(), "ssおあいうえs");

    str3.insert(str3.begin() + 1, str2.begin(), str2.begin() + 1);
    asserter->check(str3.toStr(), "あsいうえs");

    str3.insert(str3.end(), str2.begin(), str2.end());
    asserter->check(str3.toStr(), "あsいうえsssお");

    return asserter->isOk();
}


int main(int argc, char *argv[])
{
    simpletest::SimpleTestSuite suite("UTF-8 文字列テスト");
    suite.addTester(sfcr::screate(utf8_multichar_test, suite.getAsserter()));
    suite.addTester(sfcr::screate(utf8_string_test, suite.getAsserter()));
    suite.addTester(sfcr::screate(utf8_string_util_test, suite.getAsserter()));
    suite.run();
    return 0;
}
