#include <iostream>
#include <sstream>
#include <string>
#include <algorithm>

#include <functional>


#include "../simpletest.h"

#include "../utf8.h"
#include "../textarrayformat.h"

using namespace std;

struct identity_local
{
    char operator()(unsigned char t) {return static_cast<char>(t);}
};

bool utf8_charcheck_test(smart_ptr<simpletest::SimpleTestAsserter> asserter)
{
    // UTF-8の文字を認識するテスト。

    // ソースがすでにUTF8で記述されているので、そのまま簡単にだが
    // テストが行える。
    std::string tmp("UTF8");
    // 1byteの文字のみの場合
    smart_ptr<std::istream> ss(new std::stringstream(tmp));

    utakata::utf8::UTF8InputStream stream(ss);

    // 単独のreadを試す。
    asserter->check(stream.read()[0], 'U', "一文字テスト");
    // peekが問題なく働いていることのテスト
    asserter->check(stream.peek()[0], 'T', "一文字テスト"); 
    asserter->check(stream.read()[0], 'T', "一文字テスト");

    // ここの時点ではまだeofではない。
    asserter->check(stream.isEOF(), false, "末尾チェック");

    // 読み込んだ文字数を返すメンバ関数のチェック。
    asserter->check(stream.pos(), 2, "位置チェック");

    // 複数文字の読み出しのチェック
    const std::vector<unsigned char> f = stream.read(2);
    std::string t("", f.size());

    std::transform(f.begin(), f.end(), t.begin(),
                   identity_local());
    
    asserter->check(t, "F8", "複数文字チェック");

    // この時点でeofがtrueとなる
    std::vector<unsigned char> tt = stream.read();
    asserter->check(tt[0], 0xff, "huga");
    asserter->check(stream.isEOF(), true);
    asserter->check(stream.read()[0], 0xff, "hoge");
    
    return asserter->isOk();
}

bool utf8_multichar_test(smart_ptr<simpletest::SimpleTestAsserter> asserter)
{

    std::string tmp("あいうえお");
    // 1byteの文字のみの場合
    smart_ptr<std::istream> ss(new std::stringstream(tmp));

    utakata::utf8::UTF8InputStream stream(ss);

    // 複数文字の読み出しのチェック
    // 但し、ここでは全体が等しいかどうかを返すに留まる。
    std::vector<unsigned char> f = stream.read(2);
    asserter->check(true, utakata::utf8::is_utf8_all(f));

    // チェック系関数のチェック
    // わざと1byte欠落させて様子を見てみる。
    f.pop_back();
    asserter->check(false, utakata::utf8::is_utf8_all(f));

    // 欠落させても、先頭の方が生き残っていればなんとかなる。
    size_t size = 0;
    asserter->check(true, utakata::utf8::is_utf8_one(f, size));
    asserter->check(3, size);

    return asserter->isOk();
}

bool utf8_unget_test(smart_ptr<simpletest::SimpleTestAsserter> asserter)
{

    std::string tmp("あいうえお");
    // 1byteの文字のみの場合
    smart_ptr<std::istream> ss(new std::stringstream(tmp));

    utakata::utf8::UTF8InputStream stream(ss);

    // 複数文字の読み出しのチェック
    // 但し、ここでは全体が等しいかどうかを返すに留まる。
    std::vector<unsigned char> f = stream.read();

    asserter->check(utakata::utf8::generateUTF8Code(f),
                    utakata::utf8::generateUTF8Code("あ"));

    // ungetすると元に戻るので、これで正しくなるはず。
    stream.unget(f);
    asserter->check(utakata::utf8::generateUTF8Code(stream.read()),
                    utakata::utf8::generateUTF8Code("あ"));
    
    return asserter->isOk();
}

bool utf8_checker_test(smart_ptr<simpletest::SimpleTestAsserter> asserter)
{
    // 各チェック関数のチェック

    std::string tmp("059Aif");
    // 1byteの文字のみの場合
    smart_ptr<std::istream> ss(new std::stringstream(tmp));

    utakata::utf8::UTF8InputStream stream(ss);

    std::vector<unsigned char> f = stream.read();

    asserter->check(true, utakata::utf8::is_utf8_numeric(f));
    asserter->check(true, utakata::utf8::is_utf8_numeric(stream.read()));
    asserter->check(true, utakata::utf8::is_utf8_numeric(stream.read()));
    asserter->check(false, utakata::utf8::is_utf8_numeric(stream.peek()));
    asserter->check(true, utakata::utf8::is_utf8_alpha(stream.read()));
    asserter->check(true, utakata::utf8::is_utf8_ascii(stream.peek()));
    asserter->check(true, utakata::utf8::is_utf8_alpha(stream.read()));
    return asserter->isOk();
}

int main(int argc, char *argv[])
{
    simpletest::SimpleTestSuite suite("UTF-8 基本テスト");
    suite.addTester(sfcr::screate(utf8_charcheck_test, suite.getAsserter()));
    suite.addTester(sfcr::screate(utf8_multichar_test, suite.getAsserter()));
    suite.addTester(sfcr::screate(utf8_unget_test, suite.getAsserter()));
    suite.addTester(sfcr::screate(utf8_checker_test, suite.getAsserter()));
    suite.run();
    return 0;
}

