#include <vector>
#include <iostream>
#include <algorithm>
#include <string>
#include <sstream>
#include <assert.h>

#include "InputStream.h"
#include "utf8.h"
#include "smart_ptr.h"

using namespace std;;
using namespace utakata::utf8;

UTF8InputStream::UTF8InputStream() : EOF_(0xff), strm_(), pos_(0)
{
}

UTF8InputStream::UTF8InputStream(smart_ptr<std::istream> strm) : EOF_(0xff), strm_(strm),
                                                                        pos_(0)
{
}

bool UTF8InputStream::open(smart_ptr<std::istream> strm)
{
    //現在保持しているストリームと切り替える。
    // 基本的にただスワップするだけで問題ない。

    // NULLポインタではなく、問題無く開かれている場合には、次のようにして開始する。
    if (strm.isNull() != false && !strm->good()) {
        strm_ = strm;
    } else {
        return false;
    }
    return true;
}

std::vector<unsigned char> UTF8InputStream::read()
{
    // UTF-8の一文字を読みだして返す。
    // UTF-8に該当しない場合、空のvectorを返す。
    if (!strm_->eof() && !strm_->good()) {
        throw StreamException("not ready input stream");
    }

    // 最初に一文字だけ読みだして、チェックをかける。
    int c = 0;
    c = strm_->peek();
    if (c != std::istream::traits_type::eof())
    {
        // 末尾でない場合のみ、以降のチェックに入る。
        size_t size = 0;
        // 先頭1バイトが正常でなかった場合はそのまま抜ける
        if (is_utf8_first_byte(static_cast<unsigned char>(c), size))
        {
            if (size > 0)
            {
                // sizeが0より大きい場合には、この時複数バイトで文字が構成
                // されていると考えられるため、明示的に複数文字を取得する。
                std::vector<char> tmp(size, 0);
                strm_->read(&tmp[0], size);
                
                if (strm_->bad())
                {
                    // 読み取りきれなかった場合には、ストリームに一応読出せた
                    // 分を元に戻す。
                    std::for_each(tmp.rbegin(), tmp.rend(), PutBack(strm_));
                    return std::vector<unsigned char>(1, 0);
                }
                std::vector<unsigned char> rtn;
                rtn.insert(rtn.begin(), tmp.begin(), tmp.end());

                // 読みこめたので、++する。
                ++pos_;
                return rtn;
            }
        }
    }

    return std::vector<unsigned char>(1, EOF_);
}

std::vector<unsigned char> UTF8InputStream::read(int num)
{
    // 指定された文字分だけ読みだしてくる。
    // 途中で終了した場合、その文字の分だけunsigned charが減少すること
    // になっている。
    // numが0の場合、必ず空のvectorが返される。

    if (num == 0)
    {
        return std::vector<unsigned char>();
    }

    // eofの場合なら、この時点でeofが返るので、それで問題はない。
    std::vector<unsigned char> rtn = this->read();
    for (int i = 1; i < num && !strm_->eof(); ++i)
    {
        // 個数に到達するか、もしくはeofとなるまでは追加しつづける。
        std::vector<unsigned char> tmp = this->read();
        rtn.insert(rtn.end(), tmp.begin(), tmp.end());
    }

    pos_ += num;
    
    return rtn;
}

std::vector<unsigned char> UTF8InputStream::peek()
{
    // 一文字分だけ先読みする。先読みした場合、文字は戻す。
    std::vector<unsigned char> tmp = this->read();
    // 一応戻すサイズが存在する場合だけ、これを実行させることにする。
    if (tmp.size() > 0)
    {
        // 複雑な繰り返しを表現する場合には、積極的にalgorithmを利用するようにする。
        
        std::for_each(tmp.rbegin(), tmp.rend(),
                      utakata::utf8::PutBack(strm_));
        --pos_;
    }
    return tmp;
}

void UTF8InputStream::unget(const std::vector<unsigned char>& ch)
{
    // 渡されたバイト列をストリームに差し戻す。
    size_t t = 0;
    if (is_utf8_one(ch, t))
    {
        std::for_each(ch.rbegin(), ch.rend(), PutBack(strm_));
    }

    if (pos_ > 0)
    {
        --pos_;
    }
}
 
bool utakata::utf8::UTF8InputStream::isEOF() const
{
    bool ret = false;
    if (strm_->good())
    {
        ret = strm_->eof() ? true : false;
    }
    else
    {
        ret = true;
    }

    return ret;
}

//================================================================================

long utakata::utf8::generateUTF8Code(const std::vector<unsigned char>& bytes)
{
    // 1文字分のUTF8のバイト列を受け取って、コードに変換して返す。
    // 先頭の値によって、次のように値を決定することができる。
    // x = utf8の先頭バイト
    // y1〜yN = utf8の先頭バイト以降のバイト
    // N = utf8の先頭バイトを含むバイト数
    // code = (y1 & ((1 << 7) - 1)) << (6 * n-1) + (y2 & ((1 << 7) -1)) << (6 * (n - 1))...+ x & ((1 << N) -1) << (6 * N-1)
    // 先頭バイト以外は、全て先頭に10とうビットが設定されている。このビットを除いた6ビットをする。
    // つまり、末尾のバイトから順次やっていけばよい。

    std::vector<unsigned char> tmp(bytes);
    const unsigned char max_c = (1 << (sizeof(unsigned char) * 8 - 1)) - 1;
    long code = 0;

    if (tmp.empty())
    {
        return 0;
    }

    if (tmp.size() == 1)
    {
        // asciiコードは7bitなのでそこだけ切り取って返す。
        code = tmp[0] & max_c;
    }
    else
    {
        // 一時的に利用されるクラス。
        struct Lambda
        {
            unsigned char operator()(unsigned char c, int s) {
                return c << (6 * s);
            }
        };

        // サイズが1以外の場合、ここからがちと違う。
        std::vector<unsigned char>::reverse_iterator beg = tmp.rbegin(),
            end = tmp.rend() - 1;
        const unsigned char char_bit = (1 << 6) - 1;
        for (int i = 0; beg != end; ++i,++beg)
        {
            code += Lambda()((*beg & char_bit), i);
        }

        // 最後だけ、別の計算が必要になる。
        const unsigned char first_byte = (1 << ((sizeof(unsigned char) + 1) - tmp.size())) - 1;
        code += Lambda()(first_byte,tmp.size() - 1);
        
    }

    return code;
}


long utakata::utf8::generateUTF8Code(const std::string& bytes)
{
    // UTF8である一文字のstringを受け取って、先頭1文字の値を返す。

    std::string str = bytes;
    std::vector<unsigned char> tmp;
    tmp.insert(tmp.end(), str.begin(), str.end());

    // vectorにしなおしたら後は元々の関数に任せる。
    return generateUTF8Code(tmp);
}

bool utakata::utf8::is_utf8_one(const std::vector<unsigned char>& bytes, size_t& size)
{
    //渡したバイト列がUTF8の一文字に該当するかどうかを返す。
    size_t back = size;
    
    if (bytes.size() == 0)
    {
        size = back;
        return false;
    }

    size_t num = 0;

    if (!is_utf8_first_byte(bytes[0], num))
    {
        size = back;
        return false;
    }

    // そもそもbytesのサイズが足りない場合にも失敗とする。
    if (num > bytes.size())
    {
        size = back;
        return false;
    }

    // 先頭要素以外が正しければそれで問題ないとする。
    if (num > 1)
    {
        
        const CheckUTF8Byte& checker = for_each(bytes.begin() + 1, bytes.begin() + num,
                                                 CheckUTF8Byte());
        
        if (checker.good)
        {
            size = num;
        } else {
            size = back;
            return false;
        }
    } else {
        // sizeが0の場合には、この時点で1を設定するようにする。
        size = num;
    }

    // 最後まで到達した時点で成功とする。
    return true;
}

bool utakata::utf8::is_utf8_all(const std::vector<unsigned char>& bytes)
{
    // 与えられたバイト列全てがUTF-8であるかどうかを返す。
    size_t size = 0;
    std::vector<unsigned char>::const_iterator it = bytes.begin();
    while (is_utf8_one(std::vector<unsigned char>(it, bytes.end()), size)) {
        it += size;
        size = 0;
    }

    if (it == bytes.end()) {
        return true;
    }

    return false;
}

bool utakata::utf8::is_utf8_first_byte(unsigned char c, size_t& size)
{
    // UTf-8の先頭バイトであるかどうかを返す。
    // 先頭バイトである場合には、その先頭バイトを含む、一文字のサイズを返す。
    // 先頭バイトではない場合には、
    const unsigned char max_c = 1 << (sizeof(unsigned char) * 8 - 1);

    size_t back = size;

    // 最上位ビットが0である場合、これはasciiコードを指す。
    if (!(c & max_c))
    {
        size = 1;
        return true;
    }

    unsigned char first = c << 1;
    size_t num = 1;
    while (first & max_c) {
        first <<= 1;
        num += 1;
    }

    // ここまできたとき、最上位ビットは0であるはず。
    // numが5未満である場合、とりあえず正常としておくこととする。
    const unsigned char max_utf8_sequence = 5;
    if (num < max_utf8_sequence) {
        size = num;
        return true;
    }
    else if (num == 1)
    {
        // numが1の場合、何らかの理由で先頭が欠落したと見られる。
        // この場合、スキップするべきバイト数を返す。
        size = 1;
        return false;
    
    } else {
        size = back;
        return false;
    }
}


bool utakata::utf8::is_utf8_ascii(const std::vector<unsigned char>& bytes)
{
    // 一文字かつ、0x00〜0x7fの範囲であるデータであることが条件となる。
    size_t s = 0;
    bool b = is_utf8_one(bytes, s);

    if (b && s == 1)
    {
        return true;
    }
    return false;
}

bool utakata::utf8::is_utf8_numeric(const std::vector<unsigned char>& bytes)
{
    // 一文字分だけが渡されていると判断する。
    if (!is_utf8_ascii(bytes))
    {
        return false;
    }

    if (bytes[0] >= '0' && bytes[0] <= '9')
    {
        return true;
    }

    return false;
}

bool utakata::utf8::is_utf8_alpha(const std::vector<unsigned char>& bytes)
{
    // 同じく一文字であると判別する。
    // asciiのサブセットなので、先にasciiであると判別しておく。
    if (!is_utf8_ascii(bytes))
    {
        return false;
    }

    if ((bytes[0] >= 'a' && bytes[0] <= 'z') ||
        (bytes[0] >= 'A' && bytes[0] <= 'Z'))
        
    {
        return true;
    }

    return false;
}
