#ifndef _UTF8_H_
#define _UTF8_H_

#include <iostream>
#include <string>
#include <vector>
#include <exception>

#include "smart_ptr.h"
#include "InputStream.h"

namespace utakata {

    namespace utf8 {

        // inputstreamの準備が出来ていない場合に送出される例外
        class StreamException : public std::exception
        {
        public:
            StreamException(const std::string& str) : str_(str) {}
            virtual ~StreamException() throw() {}

            const char* what() throw() {
                return str_.c_str();
            }
        private:

            std::string str_;
        };

        class UTF8InputStream : public IInputStream
        {
            /**
               入力ストリームから、UTF-8のデータを指定した文字だけ読みだして
               返す。
            */

            const unsigned char EOF_;
            
        public:
            
            // 入力に利用するストリームは最初に渡される。
            // 最初に渡さない場合には、後から開けるようにしておかなけりゃならない。
            UTF8InputStream();
            UTF8InputStream(smart_ptr<std::istream> strm);
            virtual ~UTF8InputStream(){}

            bool open(smart_ptr<std::istream> strm);

            std::vector<unsigned char> read();
            std::vector<unsigned char> read(int num);

            std::vector<unsigned char> peek();
            std::vector<unsigned char> peek(int num);

            // 読み出した文字数を返す。
            size_t pos() const {return pos_;}

            // 渡したvectorをストリームに戻す。vectorは、readで取得した
            // データの並びでなければならない。つまり、read()した
            // データを戻す際に利用する。
            void unget(const std::vector<unsigned char>& ch);

            // ファイルの終端に到達しているかどうかを返す。
            // trueを返す場合、readの結果は常にEOF文字を返す。
            bool isEOF() const;
            
        private:

            smart_ptr<std::istream> strm_;
            size_t pos_;
        };


        // UTF-8のコードを表すバイト列をUTF-8のコードに変換する。
        long generateUTF8Code(const std::vector<unsigned char>& code);
        long generateUTF8Code(const std::string& ch);
        
        struct CheckUTF8Byte
        {
            // UTF8の先頭バイト以外であるかどうかをチェックする。
            const unsigned char checker;
            bool good;
            CheckUTF8Byte() : checker(0x2), good(true) {}

            template<class T>
            void operator()(const T& t) {
                // 先頭ビットが10ではない場合、チェックに失敗する。
                T tmp = t >> 6;
                if ( ((tmp & 0x3)) != checker) {
                    good = false;
                }
            }
        };

        struct PutBack
        {
            // 渡されたデータをistreamにputbackする。
            smart_ptr<std::istream> strm_;
            PutBack(smart_ptr<std::istream> strm) : strm_(strm) {}

            template<class T>
            void operator()(T t)
                {
                    strm_->putback(t);
                }
        };
        
        // 与えられたバイト列の先頭から、UTF8一文字に該当しているかどうかを返す。
        // バイト列がUTF8に該当する場合、そのバイト列のサイズを返す。
        bool is_utf8_one(const std::vector<unsigned char>& bytes, size_t& size);

        // 与えられたバイト列全てが、UTF8に該当しているかどうかを返す。
        bool is_utf8_all(const std::vector<unsigned char>& bytes);

        // UTF-8の先頭バイトとして正しいフォーマットであるかどうか。
        // 正しいフォーマットである場合、渡したバイトを含めた、一文字である
        // バイト数を返す。
        bool is_utf8_first_byte(unsigned char c, size_t& size);

        // 渡したバイト列が、0-9で判別される数値文字と一致するかどうかを返す。
        // is_asciiのサブセットを判別するものである。
        bool is_utf8_numeric(const std::vector<unsigned char>& bytes);

        // 渡したバイト列が、a-zA-Zの領域であるかどうかを調べる。
        // is_asciiのサブセットを判別するものである。
        bool is_utf8_alpha(const std::vector<unsigned char>& bytes);

        // 渡したバイト列が、0x00〜0x7fの領域であるかどうかを返す。
        // is_utf8_alpha, is_utf8_numericのスーパーセットとなる。
        bool is_utf8_ascii(const std::vector<unsigned char>& bytes);
    };
}

#endif /* _UTF8_H_ */
