#include <vector>
#include <string>
#include <functional>
#include <algorithm>

#include "utf8.h"
#include "utf8_string.h"

using namespace utakata::utf8;
using namespace utakata::utf8_string;

//================================================================================

utakata::utf8_string::UTF8Char::UTF8Char(const std::vector<unsigned char>& utf8) : utf8_bytes_(utf8), utf16_code_(0)
{
    this->utf16_code_ = utakata::utf8::generateUTF8Code(utf8_bytes_);
}

utakata::utf8_string::UTF8Char::UTF8Char(const utakata::utf8_string::UTF8Char& ch)
{
    // 単純にそれぞれの値をコピーする。
    this->utf8_bytes_ = ch.utf8_bytes_;
    utf16_code_ = ch.utf16_code_;
}

utakata::utf8_string::UTF8Char& utakata::utf8_string::UTF8Char::operator=(const utakata::utf8_string::UTF8Char& ch)
{
    // 安全なスワップを利用する。
    UTF8Char c(ch);
    swap(c);
    return *this;
}

bool utakata::utf8_string::UTF8Char::operator==(const utakata::utf8_string::UTF8Char& ch) const
{
    // 全てのUTF8Charが保持しているUTF16変換が施された
    // コードで比較を行う。実際にはlongでの比較なのでかなり速い。

    return ch.utf16_code_ == utf16_code_;
}

bool utakata::utf8_string::UTF8Char::operator!=(const utakata::utf8_string::UTF8Char& ch) const
{
    return !(*this == ch);
}

bool UTF8Char::operator<(const UTF8Char& ch) const
{
    return utf16_code_ < ch.utf16_code_;
}

bool UTF8Char::operator>(const UTF8Char& ch) const
{
    return utf16_code_ > ch.utf16_code_;
}


void utakata::utf8_string::UTF8Char::swap(utakata::utf8_string::UTF8Char& ch)
{
    // 値と標準コンテナなので、特に問題なく動作する。
    std::swap(utf16_code_, ch.utf16_code_);
    std::swap(utf8_bytes_, ch.utf8_bytes_);
}

std::string utakata::utf8_string::UTF8Char::toStr() const
{
    // 単純に置換が行える。
    std::string tmp;
    tmp.insert(tmp.end(), utf8_bytes_.begin(), utf8_bytes_.end());
    return tmp;
}

bool utakata::utf8_string::is_ascii_char(const UTF8Char& ch)
{
    // 0x7f >= ascii >= 0x00 がasciiなので、その範囲で判定を行う。
    return is_utf8_ascii(ch.getBytes()) ? true : false;
}

bool utakata::utf8_string::is_numeric(const UTF8Char& ch)
{
    return is_utf8_numeric(ch.getBytes()) ? true : false;
}

bool utakata::utf8_string::is_alpha(const UTF8Char& ch)
{
    return is_utf8_alpha(ch.getBytes()) ? true : false;
}

bool utakata::utf8_string::is_eof(const UTF8Char& ch)
{
    return ch.getBytes()[0] == 0xff ? true : false;
}

//================================================================================

utakata::utf8_string::UTF8String::UTF8String() : chars_()
{
}

utakata::utf8_string::UTF8String::UTF8String(
    const std::vector<unsigned char>& bytes) : chars_()
{
    assign(bytes);
}

utakata::utf8_string::UTF8String::UTF8String(
    const UTF8String& str) : chars_(str.chars_)
{
}

void UTF8String::assign(const std::vector<unsigned char>& bytes)
{
    // 与えられたbytesを順次utf8charに変換していく。
    // 与えられたbytesのうち、有効なデータのみを変換していく。
    std::vector<UTF8Char> chars;
    std::vector<unsigned char>::const_iterator it = bytes.begin(),
        end = bytes.end();

    while (it != end) {
        size_t i = 0;
        if (utakata::utf8::is_utf8_one(std::vector<unsigned char>(it, end), i))
        {
            // 直接生成して渡す。
            chars.push_back(std::vector<unsigned char>(it, it + i));
        }
        // 判別に失敗した場合でも、スキップするべきバイト数は
        // 渡されてくるため、とりあえず追加するだけでよい。
        it += i;
    }
    // 最後にswapするとよい。
    std::swap(chars_, chars);
}

void UTF8String::assign(const UTF8String& str)
{
    //基本的に=で渡した場合と全く同じなので、そのようにする。
    UTF8String tmp(str);
    swap(tmp);
}

UTF8String& UTF8String::operator=(const UTF8String& str)
{
    assign(str);
    return *this;
}

void UTF8String::swap(UTF8String& str)
{
    // シンプルにswapを行う。
    std::swap(chars_, str.chars_);
}

UTF8String& UTF8String::operator+=(const UTF8String& str)
{
    // 一度コピーと加算してから実際にswapさせる。
    UTF8String tmp(str);
    chars_.insert(chars_.end(), tmp.chars_.begin(), tmp.chars_.end());
    return *this;
}

UTF8String& UTF8String::operator+=(const std::vector<unsigned char>& ch)
{
    // 実際には文字を設定するのに利用される。
    // また、これを定義しておくことで、streamからの結果を直接設定することができる。
    UTF8String tmp(ch);
    chars_.insert(chars_.end(), tmp.chars_.begin(), tmp.chars_.end());
    return *this;
}

UTF8String& UTF8String::operator+=(const std::string& str)
{
    // 実際には文字を設定するのに利用される。
    // また、これを定義しておくことで、streamからの結果を直接設定することができる。
    std::vector<unsigned char> t;
    std::string::const_iterator begin = str.begin(), end = str.end();
    for (; begin != end; ++begin)
    {
        t.push_back(*begin);
    }

    UTF8String tmp(t);
    chars_.insert(chars_.end(), tmp.chars_.begin(), tmp.chars_.end());
    return *this;
}

UTF8String& UTF8String::operator+=(const UTF8Char& ch)
{
    // 実際には文字を設定するのに利用される。
    // また、これを定義しておくことで、streamからの結果を直接設定することができる。
    UTF8String tmp(ch.getBytes());
    chars_.insert(chars_.end(), tmp.chars_.begin(), tmp.chars_.end());
    return *this;
}


std::string UTF8String::toStr() const
{
    // 文字列に変換する。
    std::string ret;
    std::for_each(chars_.begin(), chars_.end(),
                  UTF8StringToString<UTF8Char>(ret));
    return ret;
}

void UTF8String::insert(UTF8String::utf8iterator it, UTF8String::utf8iterator begin,
                        UTF8String::utf8iterator last)
{
    // 渡されたのは実際にはvectorのイテレータなので、
    // そのままvectorの実装に任せることができる。
    chars_.insert(it, begin, last);
}

void UTF8String::insert(UTF8String::utf8iterator it,
                        UTF8String::const_utf8iterator begin,
                        UTF8String::const_utf8iterator last)
{
    // 渡されたのがconstであるかどうかというだけの違いであるため、
    // そのまま渡すことができる。
    chars_.insert(it, begin, last);
}

//================================================================================

utakata::utf8_string::UTF8String utakata::utf8_string::substring(
    const utakata::utf8_string::UTF8String& str, size_t begin, size_t end)
{
    if (end == 0)
    {
        // 先頭から末尾までを取得する。
        UTF8String s;
        s.insert(s.begin(), str.begin() + begin, str.end());
        return s;
    }
    else if (begin <= end)
    {
        UTF8String s;
        s.insert(s.begin(), str.begin() + begin, str.begin() + (begin + end));
        return s;
    }
    else
    {
        throw range_error("out of range in utakata::utf8_string::substring");
    }
}

utakata::utf8_string::UTF8String utakata::utf8_string::convert(
    const std::string& str)
{
    UTF8String tmp;
    tmp += str;
    return tmp;
}

UTF8String utakata::utf8_string::operator+(const UTF8String& lh, const UTF8String& rh)
{
    // 双方をコピーして加算して返す。凄い負荷が高い。
    UTF8String str(lh);
    str += rh;
    return str;
}

UTF8String utakata::utf8_string::operator+(const UTF8String& lh, const UTF8Char& rh)
{
    // 文字と加算する。
    UTF8String tmp(lh);
    tmp += rh.getBytes();
    return tmp;
}

UTF8String utakata::utf8_string::operator+(const UTF8Char& lh, const UTF8String& rh)
{
    UTF8String tmp(lh.getBytes());
    tmp += rh;
    return tmp;
}
