#ifndef _UTF8_STRING_H_
#define _UTF8_STRING_H_

#include <vector>
#include <functional>
#include <string>
#include <exception>


// UTF8のバイト列、あるいは文字列から、UTF8の一文字単位で保持を行う
// CUTF8Charの作成と、CUTF8CharのコンテナであるCUTF8Stringを定義する。


namespace utakata {

    namespace utf8_string {

        class range_error : public std::exception
        {
        public:
            range_error(const std::string& str) : str_(str) {}
            virtual ~range_error() throw() {}

            const char* what() throw() {
                return str_.c_str();
            }
        private:

            std::string str_;
        };
    
        class UTF8Char
        {
            /**
               UTF8と判断されたバイト列を受け取り、実際の値を返す。
               ここで言う実際の値とは、UTF-8で定まっているバイト列の表現から、
               固定とされる表現部分を削除し、有効とされる値を結合した値を指す。
               この値は実際のUTF16のコードと同一であり、完全なマッピングが可能
               となる。

               110xxxxx 10yyyyyy --> xxxxxyyyyyy = 11bit
               1110xxxx 10yyyyyy 10zzzzzz --> xxxxyyyyyyzzzzzz = 16bit
               11110xxx 10yyyyyy 10zzzzzz 10vvvvvv --> xxxyyyyyyzzzzzzvvvvvv = 21bit
               ※策定仕様では、5byte以上となるUTF-8は異常であるとされるため、21bitが表現
               できればよい。

               21bitを十分に表現することができる整数型は、longのみであるため、必ずlongとして
               返す。
               intに暗黙のキャストをすることができるが、intは16bitまでしか保持義務が無いため、
               原則としてCUTF8Charはlongで扱う。
            */

        public:
            // バイト列を必ず渡す必要がある。
            UTF8Char(const std::vector<unsigned char>& utf8);
            virtual ~UTF8Char(){}

            // コピーコンストラクタを宣言したため、同時にoperator=を宣言する。
            UTF8Char(const UTF8Char& ch);
            UTF8Char& operator=(const UTF8Char& ch);

            // 各演算子のオーバーロード。
            // <=と>=は、それぞれ>の結果のnot、<の結果のnotを
            // 用いることで実装できるため、この二つだけ供えていれば問題ない。
            // また、比較には内部で保持しているutf16_code_を用いるため、
            // 速度面で気にする必要は無い。
            bool operator==(const UTF8Char& code) const;
            bool operator<(const UTF8Char& code) const;
            bool operator>(const UTF8Char& code) const;
            bool operator!=(const UTF8Char& code) const;

            // UTF8->UTF16に変換したコードを返す。
            long toUTF16Code() const {return this->utf16_code_;}

            // 実際のバイト列を返す。
            const std::vector<unsigned char>& getBytes() const {return utf8_bytes_;}

            // 文字列として変換して返す。
            std::string toStr() const;

            // 渡されたUTF8Charの内部と交換する。
            void swap(UTF8Char& ch);

        private:

            // utf8を表すバイト列
            std::vector<unsigned char> utf8_bytes_;
            // 渡されたバイト列を実際のコードにしたもの
            long utf16_code_;
        };

        // 渡されたUTF8Charがasciiコードの範囲内に収まっているかどうかを返す。
        bool is_ascii_char(const UTF8Char& ch);

        // 渡されたUTF8Charが、それぞれ数値、アルファベットであるかどうか
        // を返す。これらはutf8モジュールのユーティリティ関数によって実装される。
        bool is_numeric(const UTF8Char& ch);
        bool is_alpha(const UTF8Char& ch);

        // UTF8では先頭1バイトが0xffになることはありえないので、
        // 先頭1バイトが0xffの場合には、これは終端記号であるとした。
        bool is_eof(const UTF8Char& ch);

        //================================================================================

        class UTF8String
        {
            // UTF8の文字列を保持するためのコンテナ。
            // 相互での比較などもサポートし、utakataの内部全般で
            // 標準として使用することが可能なものとし、かつ標準コンテナと
            // ほぼ同様の処理を行うことができるように定義されている。
            // というか内部では単純なvectorで定義されているため、
            // typedefされた型を指定するだけで、イテレータとして利用可能。

        public:

            // イテレータとして簡単に利用するためのtypedefマクロ
            typedef std::vector<UTF8Char>::iterator utf8iterator;
            typedef std::vector<UTF8Char>::const_iterator const_utf8iterator;

        public:
            
            UTF8String();

            UTF8String(const std::vector<unsigned char>& bytes);
            UTF8String(const UTF8String& str);
            virtual ~UTF8String(){}

            // 実体に代入する。代入が行われなかった場合、元のデータ
            // は保存される。
            void assign(const std::vector<unsigned char>& bytes);
            void assign(const UTF8String& str);

            // iteratorを取得する。
            const_utf8iterator begin() const {return chars_.begin();}
            utf8iterator begin() {return chars_.begin();}

            // 末尾のイテレータを取得する。
            const_utf8iterator end() const {return chars_.end();}
            utf8iterator end() {return chars_.end();}

            // []をオーバーロードする。これは常に境界のチェックを行う。
            const UTF8Char& operator[](size_t t) const {return chars_.at(t);}
            UTF8Char& operator[](size_t t) {
                return const_cast<UTF8Char&>(
                    static_cast<const UTF8String*>(this)->chars_[t]);}

            // サイズを取得する。
            size_t size() const {return chars_.size();}
            // 空かどうかを調べる。
            bool empty() const {return size() == 0 ? true : false;}

            // 安全なswapと共に提供する。
            UTF8String& operator=(const UTF8String& str);
            void swap(UTF8String& str);

            // 加算のみをサポートする。
            UTF8String& operator+=(const UTF8String& str);
            UTF8String& operator+=(const std::vector<unsigned char>& ch);
            UTF8String& operator+=(const UTF8Char& ch);
            UTF8String& operator+=(const std::string& str);

            // std::stringへ変換する。ただし、std::string上でのinsertなどは
            // 保証できない。
            std::string toStr() const;

            // 渡したイテレータの範囲を追加する
            // beginとlastは同一のUTF8Stringから取得されたものであること。
            void insert(utf8iterator it, utf8iterator begin, utf8iterator last);
            void insert(utf8iterator it, const_utf8iterator begin,
                        const_utf8iterator last);

        private:

            std::vector<UTF8Char> chars_;
        };

        UTF8String operator+(const UTF8String& lh, const UTF8String& rh);
        UTF8String operator+(const UTF8String& lh, const UTF8Char& rh);
        UTF8String operator+(const UTF8Char& lh, const UTF8String& rh);

        // std::stringから直接utf8stringをを作成する。
        UTF8String convert(const std::string& str);

        // substringの実装を行う。
        // [begin, end)までの文字を文字列として返す。
        // endが渡されないか、0が渡された場合、beginから末尾までが返される。
        UTF8String substring(const UTF8String& str, size_t begin, size_t end = 0);

        template<class T>
        class UTF8StringToString : public std::unary_function<T, void>
        {
            // イテレータを受け取り、std::stringを初期化引数で渡されたものに
            // コピーしていく。
        public:

            UTF8StringToString(std::string& str) : ret_(str) {}
            void operator()(const T& it) {
                ret_ += it.toStr();
            }
            
        private:
            
            std::string& ret_;
        };
    };

};

#endif /* _UTF8_STRING_H_ */
